# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import ipaddress
from typing import Optional, List

from kubemarine.core import utils, log
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup
from kubemarine.plugins.manifest import Processor, EnrichmentFunction, Manifest


def check_job_for_nginx(cluster: KubernetesCluster) -> None:
    first_control_plane = cluster.nodes['control-plane'].get_first_member()
    version = cluster.inventory['plugins']['nginx-ingress-controller']['version'].replace('v', '.').split('.')

    major_version = int(version[1])
    minor_version = int(version[2])

    check_jobs = first_control_plane.sudo(f"kubectl get jobs -n ingress-nginx")
    if list(check_jobs.values())[0].stderr == "" and major_version >= 1 and minor_version >= 4:
        cluster.log.debug('Delete old jobs for nginx')
        first_control_plane.sudo(f"sudo kubectl delete job --all -n ingress-nginx")
    else:
        cluster.log.debug('There are no jobs to delete')


def enrich_inventory(inventory: dict, cluster: KubernetesCluster) -> dict:
    if not inventory["plugins"]["nginx-ingress-controller"]["install"]:
        return inventory

    if inventory["plugins"]["nginx-ingress-controller"].get('custom_headers'):
        if not inventory["plugins"]["nginx-ingress-controller"].get('config_map'):
            inventory["plugins"]["nginx-ingress-controller"]['config_map'] = {}
        if not inventory["plugins"]["nginx-ingress-controller"]['config_map'].get('proxy-set-headers'):
            inventory["plugins"]["nginx-ingress-controller"]['config_map']['proxy-set-headers'] = 'ingress-nginx/custom-headers'

    # if user defined resources himself, we should use them as is, instead of merging with our defaults
    raw_controller = cluster.raw_inventory.get("plugins", {}).get("nginx-ingress-controller", {}).get("controller", {})
    if "resources" in raw_controller:
        inventory["plugins"]["nginx-ingress-controller"]["controller"]["resources"] = raw_controller["resources"]
    raw_webhook = cluster.raw_inventory.get("plugins", {}).get("nginx-ingress-controller", {}).get("webhook", {})
    if "resources" in raw_webhook:
        inventory["plugins"]["nginx-ingress-controller"]["webhook"]["resources"] = raw_webhook["resources"]

    return inventory


def cert_renew_enrichment(inventory: dict, cluster: KubernetesCluster) -> dict:
    # check that renewal is required for nginx
    if cluster.context.get('initial_procedure') != 'cert_renew' \
            or not cluster.procedure_inventory.get("nginx-ingress-controller"):
        return inventory

    nginx_plugin = inventory["plugins"]["nginx-ingress-controller"]

    # check that renewal is possible
    if not nginx_plugin["install"]:
        raise Exception("Certificates can not be renewed for nginx plugin since it is not installed")

    # update certificates in inventory
    nginx_plugin["controller"]["ssl"]["default-certificate"] = cluster.procedure_inventory["nginx-ingress-controller"]

    return inventory


def finalize_inventory(cluster: KubernetesCluster, inventory_to_finalize: dict) -> dict:
    # check that renewal is required for nginx
    if cluster.context.get('initial_procedure') != 'cert_renew' \
            or not cluster.procedure_inventory.get("nginx-ingress-controller"):
        return inventory_to_finalize

    if not inventory_to_finalize["plugins"].get("nginx-ingress-controller"):
        inventory_to_finalize["plugins"]["nginx-ingress-controller"] = {}

    if not inventory_to_finalize["plugins"]["nginx-ingress-controller"].get("controller"):
        inventory_to_finalize["plugins"]["nginx-ingress-controller"]["controller"] = {}

    if not inventory_to_finalize["plugins"]["nginx-ingress-controller"]["controller"].get("ssl"):
        inventory_to_finalize["plugins"]["nginx-ingress-controller"]["controller"]["ssl"] = {}

    nginx_plugin = inventory_to_finalize["plugins"]["nginx-ingress-controller"]
    nginx_plugin["controller"]["ssl"]["default-certificate"] = cluster.procedure_inventory["nginx-ingress-controller"]

    return inventory_to_finalize


def manage_custom_certificate(cluster: KubernetesCluster) -> None:
    if not cluster.inventory["plugins"]["nginx-ingress-controller"]["controller"]["ssl"].get("default-certificate"):
        cluster.log.debug("No custom default ingress certificate specified, skipping...")
        return

    base_path = "/etc/kubernetes/custom-certs"
    certificate_path = base_path + "/cert"
    private_key_path = base_path + "/key"
    secret_name = "default-ingress-cert"
    secret_namespace = "kube-system"

    first_control_plane = cluster.nodes["control-plane"].get_first_member()
    default_cert = cluster.inventory["plugins"]["nginx-ingress-controller"]["controller"]["ssl"]["default-certificate"]

    # first, we need to load cert and key files to first control-plane to known locations
    first_control_plane.sudo(f"mkdir -p {base_path}")
    try:
        first_control_plane.call(put_custom_certificate,
                          default_cert=default_cert,
                          crt_path=certificate_path,
                          key_path=private_key_path)

        # second, we need to validate cert and key using openssl
        first_control_plane.call(verify_certificate_and_key, crt_path=certificate_path, key_path=private_key_path)

        # third, we need to create tls secret under well-known name
        # this certificate is already configured to be used by controller
        first_control_plane.call(create_tls_secret,
                          crt_path=certificate_path,
                          key_path=private_key_path,
                          name=secret_name,
                          namespace=secret_namespace)
    finally:
        # fourth, we need to remove base path dir
        first_control_plane.sudo(f"rm -rf {base_path}")


def put_custom_certificate(first_control_plane: NodeGroup, default_cert: dict, crt_path: str, key_path: str) -> None:
    if default_cert.get("data"):
        cert = io.StringIO(default_cert["data"]["cert"])
        key = io.StringIO(default_cert["data"]["key"])
    else:
        cert = io.StringIO(utils.read_external(default_cert["paths"]["cert"]))
        key = io.StringIO(utils.read_external(default_cert["paths"]["key"]))

    first_control_plane.put(cert, crt_path, sudo=True)
    first_control_plane.put(key, key_path, sudo=True)


def verify_certificate_and_key(first_control_plane: NodeGroup, crt_path: str, key_path: str) -> None:
    crt_md5 = first_control_plane.sudo(f"openssl x509 -noout -modulus -in {crt_path} | openssl md5").get_simple_out()
    key_md5 = first_control_plane.sudo(f"openssl rsa -noout -modulus -in {key_path} | openssl md5").get_simple_out()
    if crt_md5 != key_md5:
        raise Exception("Custom default ingress certificate and key are not compatible!")


def create_tls_secret(first_control_plane: NodeGroup, crt_path: str, key_path: str, name: str, namespace: str) -> None:
    first_control_plane.sudo(f"kubectl create secret tls {name} --key {key_path} --cert {crt_path} -n {namespace} "
                      f"--dry-run -o yaml | sudo kubectl apply -f -", timeout=300)


class IngressNginxManifestProcessor(Processor):
    def __init__(self, logger: log.VerboseLogger, inventory: dict,
                 original_yaml_path: Optional[str] = None, destination_name: Optional[str] = None) -> None:
        super().__init__(logger, inventory, 'nginx-ingress-controller', original_yaml_path, destination_name)

    def get_known_objects(self) -> List[str]:
        return [
            "Namespace_ingress-nginx",
            "ServiceAccount_ingress-nginx",
            "ServiceAccount_ingress-nginx-admission",
            "Role_ingress-nginx",
            "Role_ingress-nginx-admission",
            "ClusterRole_ingress-nginx",
            "ClusterRole_ingress-nginx-admission",
            "RoleBinding_ingress-nginx",
            "RoleBinding_ingress-nginx-admission",
            "ClusterRoleBinding_ingress-nginx",
            "ClusterRoleBinding_ingress-nginx-admission",
            "ConfigMap_ingress-nginx-controller",
            "Service_ingress-nginx-controller",
            "Service_ingress-nginx-controller-admission",
            "Deployment_ingress-nginx-controller",
            "Job_ingress-nginx-admission-create",
            "Job_ingress-nginx-admission-patch",
            "IngressClass_nginx",
            "ValidatingWebhookConfiguration_ingress-nginx-admission",
        ]

    def get_enrichment_functions(self) -> List[EnrichmentFunction]:
        return [
            self.enrich_namespace_ingress_nginx,
            self.enrich_configmap_ingress_nginx_controller,
            self.add_configmap_ingress_nginx_controller,
            self.enrich_deployment_ingress_nginx_controller,
            self.enrich_ingressclass_nginx,
            self.enrich_job_ingress_nginx_admission_create,
            self.enrich_job_ingress_nginx_admission_patch,
            self.enrich_service_ingress_nginx_controller,
        ]

    def enrich_namespace_ingress_nginx(self, manifest: Manifest) -> None:
        key = "Namespace_ingress-nginx"
        rbac = self.inventory['rbac']
        if rbac['admission'] == 'pss' and rbac['pss']['pod-security'] == 'enabled' \
                and rbac['pss']['defaults']['enforce'] != 'privileged':
            self.assign_default_pss_labels(manifest, key, 'privileged')

    def enrich_configmap_ingress_nginx_controller(self, manifest: Manifest) -> None:
        key = "ConfigMap_ingress-nginx-controller"
        config_map = self.inventory['plugins']['nginx-ingress-controller'].get('config_map')
        if config_map:
            source_yaml = manifest.get_obj(key, patch=True)
            data: dict = source_yaml['data']
            data.update(config_map)
            self.log.verbose(f"The {key} has been patched in 'data' "
                             f"with the data from 'plugins.nginx-ingress-controller.config_map'")

    def add_configmap_ingress_nginx_controller(self, manifest: Manifest) -> None:
        custom_headers = self.inventory['plugins']['nginx-ingress-controller'].get('custom_headers')
        if custom_headers:
            custom_headers_cm = dict(CUSTOM_HEADERS_CM)
            custom_headers_cm['data'] = custom_headers
            # Insert custom-headers ConfigMap before ingress-nginx-controller ConfigMap
            ingres_nginx_cm = manifest.key_index("ConfigMap_ingress-nginx-controller")
            self.include(manifest, ingres_nginx_cm, custom_headers_cm)
            self.log.verbose(f"The {manifest.obj_key(custom_headers_cm)} has been patched in 'data' "
                             f"with the data from 'plugins.nginx-ingress-controller.custom_headers'")

    def enrich_deployment_ingress_nginx_controller(self, manifest: Manifest) -> None:
        key = "Deployment_ingress-nginx-controller"
        source_yaml = manifest.get_obj(key, patch=True)

        container_pos, container = self.find_container_for_patch(
            manifest, key, container_name='controller', is_init_container=False)
        self.enrich_deamonset_ingress_nginx_controller_container(container_pos, container)

        self.enrich_image_for_container(manifest, key,
            plugin_service='controller', container_name='controller', is_init_container=False)

        self.enrich_resources_for_container(manifest, key, container_name='controller', plugin_service="controller")
        self.enrich_node_selector(manifest, key, plugin_service='controller')
        self.enrich_tolerations(manifest, key, plugin_service='controller')

        # Patch kind in the last step to avoid sudden key change in log messages
        source_yaml['kind'] = 'DaemonSet'
        self.log.verbose(f"The {key} has been patched in 'kind' with 'DaemonSet'")

    def enrich_deamonset_ingress_nginx_controller_container(self, container_pos: int, container: dict) -> None:
        key = "Deployment_ingress-nginx-controller"
        container_args = container['args']
        for i, arg in enumerate(container_args):
            if arg.startswith('--publish-service='):
                del container_args[i]
                self.log.verbose(f"The {arg!r} argument has been removed from "
                                 f"'spec.template.spec.containers.[{container_pos}].args' in the {key}")
                break
        else:
            raise Exception("Failed to find '--publish-service' argument in ingress-nginx-controller container specification.")

        extra_args: List[tuple] = [
            ('--watch-ingress-without-class=', 'true')
        ]
        ssl_options = self.inventory['plugins']['nginx-ingress-controller']['controller']['ssl']
        additional_args = self.inventory['plugins']['nginx-ingress-controller']['controller'].get('args')

        if additional_args:
            for arg in additional_args:
                pars_arg = arg.split('=')
                for i, container_arg in enumerate(container_args):
                    if container_arg.startswith(pars_arg[0]):
                       raise Exception(
                           f"{pars_arg[0]!r} argument is already defined in ingress-nginx-controller container specification.")
                else:
                    container_args.append(arg)
                    self.log.verbose(f"The {arg!r} argument has been added to "
                                     f"'spec.template.spec.containers.[{container_pos}].args' in the {key}")

        if ssl_options['enableSslPassthrough']:
            extra_args.append(('--enable-ssl-passthrough',))
        if ssl_options.get('default-certificate'):
            extra_args.append(('--default-ssl-certificate=', 'kube-system/default-ingress-cert'))

        for extra_arg in extra_args:
            for i, arg in enumerate(container_args):
                if arg.startswith(extra_arg[0]):
                    raise Exception(
                        f"{extra_arg[0]!r} argument is already defined in ingress-nginx-controller container specification.")
            else:
                arg = extra_arg[0]
                if len(extra_arg) > 1:
                    arg += extra_arg[1]
                container_args.append(arg)
                self.log.verbose(f"The {arg!r} argument has been added to "
                                 f"'spec.template.spec.containers.[{container_pos}].args' in the {key}")

        container['ports'] = self.inventory['plugins']['nginx-ingress-controller']['ports']
        self.log.verbose(f"The {key} has been patched in 'spec.template.spec.containers.[{container_pos}].ports' "
                         f"with the data from 'plugins.nginx-ingress-controller.ports'")

    def enrich_ingressclass_nginx(self, manifest: Manifest) -> None:
        key = "IngressClass_nginx"
        source_yaml = manifest.get_obj(key, patch=True)
        source_yaml['metadata'].setdefault('annotations', {})['ingressclass.kubernetes.io/is-default-class'] = 'true'
        self.log.verbose(f"The {key} has been patched in 'metadata.annotations' "
                         f"with 'ingressclass.kubernetes.io/is-default-class: true'")

    def enrich_job_ingress_nginx_admission_create(self, manifest: Manifest) -> None:
        key = "Job_ingress-nginx-admission-create"
        self.enrich_image_for_container(manifest, key,
            plugin_service='webhook', container_name='create', is_init_container=False)

        self.enrich_resources_for_container(manifest, key, container_name='create', plugin_service="webhook")

    def enrich_job_ingress_nginx_admission_patch(self, manifest: Manifest) -> None:
        key = "Job_ingress-nginx-admission-patch"
        self.enrich_image_for_container(manifest, key,
            plugin_service='webhook', container_name='patch', is_init_container=False)

        self.enrich_resources_for_container(manifest, key, container_name='patch', plugin_service="webhook")

    def enrich_service_ingress_nginx_controller(self, manifest: Manifest) -> None:
        # The method needs some rework in case of dual stack support
        key = "Service_ingress-nginx-controller"
        ip = self.inventory['services']['kubeadm']['networking']['serviceSubnet'].split('/')[0]
        if type(ipaddress.ip_address(ip)) is ipaddress.IPv6Address:
            source_yaml = manifest.get_obj(key, patch=True)
            source_yaml['spec']['ipFamilies'] = ['IPv6']
            self.log.verbose(f"The {key} has been patched in 'spec.ipFamilies' with 'IPv6'")


class V1_2_X_IngressNginxManifestProcessor(IngressNginxManifestProcessor):
    def get_enrichment_functions(self) -> List[EnrichmentFunction]:
        enrichment_functions = super().get_enrichment_functions()
        enrichment_functions.extend([
            self.exclude_webhook_resources,
            self.enrich_role_ingress_nginx,
        ])
        return enrichment_functions

    def enrich_configmap_ingress_nginx_controller(self, manifest: Manifest) -> None:
        key = "ConfigMap_ingress-nginx-controller"
        source_yaml = manifest.get_obj(key, patch=True)
        # For some reason, we took manifest for Digital Ocean, removed use-proxy-protocol: "true" property,
        # but left service.beta.kubernetes.io/do-loadbalancer-enable-proxy-protocol: "true" annotation
        # in ingress-nginx-controller Service.
        # The behaviour is left as-is for compatibility.
        # For v.1.4.0 we took the default manifest which lacks of the mentioned properties.
        del source_yaml['data']['use-proxy-protocol']
        self.log.verbose(f"The 'use-proxy-protocol' property has been removed from 'data' in the {key}")
        super().enrich_configmap_ingress_nginx_controller(manifest)

    def enrich_deamonset_ingress_nginx_controller_container(self, container_pos: int, container: dict) -> None:
        key = "Deployment_ingress-nginx-controller"
        container_args = container['args']
        webhook_args_remove = [
            '--validating-webhook=',
            '--validating-webhook-certificate=',
            '--validating-webhook-key='
        ]
        for arg_remove in webhook_args_remove:
            for i, arg in enumerate(container_args):
                if arg.startswith(arg_remove):
                    del container_args[i]
                    self.log.verbose(f"The {arg!r} argument has been removed from "
                                     f"'spec.template.spec.containers.[{container_pos}].args' in the {key}")
                    break

        del container['volumeMounts']
        self.log.verbose(f"The 'volumeMounts' property has been removed "
                         f"from 'spec.template.spec.containers.[{container_pos}]' in the {key}")

        super().enrich_deamonset_ingress_nginx_controller_container(container_pos, container)

    def enrich_job_ingress_nginx_admission_create(self, manifest: Manifest) -> None:
        return

    def enrich_job_ingress_nginx_admission_patch(self, manifest: Manifest) -> None:
        return

    def exclude_webhook_resources(self, manifest: Manifest) -> None:
        webhook_resources = [
            "ServiceAccount_ingress-nginx-admission",
            "Role_ingress-nginx-admission",
            "ClusterRole_ingress-nginx-admission",
            "RoleBinding_ingress-nginx-admission",
            "ClusterRoleBinding_ingress-nginx-admission",
            "Service_ingress-nginx-controller-admission",
            "Job_ingress-nginx-admission-create",
            "Job_ingress-nginx-admission-patch",
            "ValidatingWebhookConfiguration_ingress-nginx-admission",
        ]
        for key in webhook_resources:
            self.exclude(manifest, key)

    def enrich_role_ingress_nginx(self, manifest: Manifest) -> None:
        key = "Role_ingress-nginx"
        source_yaml = manifest.get_obj(key, patch=True)
        # TODO patch only if psp is enabled?
        api_list = source_yaml['rules']
        api_list.append(psp_ingress_nginx)
        self.log.verbose(f"The {key} has been patched in 'rules' with {psp_ingress_nginx}")


def get_ingress_nginx_manifest_processor(logger: log.VerboseLogger, inventory: dict,
                                         yaml_path: Optional[str] = None, destination: Optional[str] = None) -> Processor:
    version: str = inventory['plugins']['nginx-ingress-controller']['version']
    kwargs = {'original_yaml_path': yaml_path, 'destination_name': destination}
    if utils.minor_version(version) == 'v1.2':
        return V1_2_X_IngressNginxManifestProcessor(logger, inventory, **kwargs)

    return IngressNginxManifestProcessor(logger, inventory, **kwargs)


CUSTOM_HEADERS_CM = {
    "apiVersion": "v1",
    "kind": "ConfigMap",
    "metadata": {
        "name": "custom-headers",
        "namespace": "ingress-nginx"
    }
}

psp_ingress_nginx = {
    "apiGroups": ["extensions"],
    "resources": ["podsecuritypolicies"],
    "verbs":     ["use"],
    "resourceNames": ["oob-host-network-psp"]
}



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import uuid
import re
from typing import Dict, Any, List, Optional

import ruamel.yaml
import yaml
from jinja2 import Template

from kubemarine import kubernetes
from kubemarine.core import utils
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup, RunnersGroupResult
from kubemarine.core.yaml_merger import default_merger

privileged_policy_filename = "privileged.yaml"
policies_file_path = "./resources/psp/"
tmp_filepath_pattern = "/tmp/%s"

admission_template = "./templates/admission.yaml.j2"
admission_dir = "/etc/kubernetes/pki"
admission_path = "%s/admission.yaml" % admission_dir

psp_list_option = "psp-list"
roles_list_option = "roles-list"
bindings_list_option = "bindings-list"

provided_oob_policies = ["default", "host-network", "anyuid"]

valid_modes = ['enforce', 'audit', 'warn']
valid_versions_templ = r"^v1\.\d{1,2}$"

baseline_plugins = {"kubernetes-dashboard": "kubernetes-dashboard"} 
privileged_plugins = {"nginx-ingress-controller": "ingress-nginx", 
                      "local-path-provisioner": "local-path-storage"}

loaded_oob_policies = {}

# TODO: When KubeMarine is not support Kubernetes version lower than 1.25, the PSP implementation code should be deleted 

def enrich_inventory_psp(inventory: dict, _: KubernetesCluster) -> dict:
    global loaded_oob_policies
    loaded_oob_policies = load_oob_policies_files()

    # validate custom
    custom_policies = inventory["rbac"]["psp"]["custom-policies"]
    verify_custom(custom_policies)

    # do not perform enrichment if security disabled
    if not is_security_enabled(inventory):
        return inventory

    # if security enabled, then add PodSecurityPolicy admission plugin
    enabled_admissions = inventory["services"]["kubeadm"]["apiServer"]["extraArgs"]["enable-admission-plugins"]
    if 'PodSecurityPolicy' not in enabled_admissions:
        enabled_admissions = "%s,PodSecurityPolicy" % enabled_admissions
        inventory["services"]["kubeadm"]["apiServer"]["extraArgs"]["enable-admission-plugins"] = enabled_admissions
        
    return inventory


def enrich_inventory_pss(inventory: dict, _: KubernetesCluster) -> dict:
    if not is_security_enabled(inventory):
        return inventory
    # check flags, enforce and logs parameters
    minor_version = int(inventory["services"]["kubeadm"]["kubernetesVersion"].split('.')[1])
    if minor_version < 23:
        raise Exception("PSS is not supported properly in Kubernetes version before v1.23")
    for item in inventory["rbac"]["pss"]["defaults"]:
        if item.endswith("version"):
            verify_version(item, inventory["rbac"]["pss"]["defaults"][item], minor_version)

    # add extraArgs to kube-apiserver config
    extra_args = inventory["services"]["kubeadm"]["apiServer"]["extraArgs"]
    if minor_version <= 27:
        enabled_admissions = extra_args.get("feature-gates")
        if enabled_admissions:
            if 'PodSecurity=true' not in enabled_admissions:
                enabled_admissions = "%s,PodSecurity=true" % enabled_admissions
        else:
            enabled_admissions = "PodSecurity=true"

        extra_args["feature-gates"] = enabled_admissions

    extra_args["admission-control-config-file"] = admission_path

    return inventory


def enrich_inventory(inventory: dict, _: KubernetesCluster) -> dict:
    admission_impl = inventory['rbac']['admission']
    if admission_impl == "psp":
        return enrich_inventory_psp(inventory, _)
    elif admission_impl == "pss":
        return enrich_inventory_pss(inventory, _)

    return inventory


def manage_psp_enrichment(inventory: dict, cluster: KubernetesCluster) -> dict:
    minor_version = int(inventory["services"]["kubeadm"]["kubernetesVersion"].split('.')[1])
    if minor_version >= 25:
        raise Exception("PSP is not supported in Kubernetes version higher than v1.24")
    if cluster.context.get('initial_procedure') != 'manage_psp':
        return inventory

    procedure_config = cluster.procedure_inventory["psp"]
    current_config = cluster.inventory["rbac"]["psp"]

    # validate added custom
    custom_add_policies = procedure_config.get("add-policies", {})
    verify_custom(custom_add_policies)

    # validate deleted custom
    custom_delete_policies = procedure_config.get("delete-policies", {})
    verify_custom(custom_delete_policies)

    # forbid managing OOB if security will be disabled
    current_security_state = current_config["pod-security"]
    final_security_state = procedure_config.get("pod-security", current_security_state)
    if final_security_state == "disabled" and procedure_config.get("oob-policies"):
        raise Exception("OOB policies can not be configured when security is disabled")

    return inventory


def verify_custom(custom_scope: Dict[str, List[dict]]) -> None:
    psp_list = custom_scope.get(psp_list_option, None)
    if psp_list:
        verify_custom_list(psp_list, "PSP")

    roles_list = custom_scope.get(roles_list_option, None)
    if roles_list:
        verify_custom_list(roles_list, "role")

    bindings_list = custom_scope.get(bindings_list_option, None)
    if bindings_list:
        verify_custom_list(bindings_list, "binding")


def verify_custom_list(custom_list: List[dict], type: str) -> None:
    for item in custom_list:
        # forbid using 'oob-' prefix in order to avoid conflicts of our policies and users policies
        if item["metadata"]["name"].startswith("oob-"):
            raise Exception("Name %s is not allowed for custom %s" % (item["metadata"]["name"], type))


def verify_version(owner: str, version: str, minor_version_cfg: int) -> None:
    # check Kubernetes version and admission config matching
    if version != "latest":
        result = re.match(valid_versions_templ, version)
        if result is None:
            raise Exception("incorrect Kubernetes version %s, valid version(for example): v1.23" % owner)
        minor_version = int(version.split('.')[1])
        if minor_version > minor_version_cfg:
            raise Exception("%s version must not be higher than Kubernetes version" % owner)


def finalize_inventory_psp(cluster: KubernetesCluster, inventory_to_finalize: dict) -> dict:
    if cluster.context.get('initial_procedure') != 'manage_psp':
        return inventory_to_finalize
    procedure_config = cluster.procedure_inventory["psp"]

    if "rbac" not in inventory_to_finalize:
        inventory_to_finalize["rbac"] = {}
    if "psp" not in inventory_to_finalize["rbac"]:
        inventory_to_finalize["rbac"]["psp"] = {}
    current_config = inventory_to_finalize["rbac"]["psp"]

    # Perform custom-policies lists changes.
    # Perform changes only if there are any "custom-policies" or "add-policies" in inventory,
    # do not perform changes if only "delete-policies" defined, there is nothing to delete from inventory in this case.
    adding_custom_policies = procedure_config.get("add-policies", {})
    deleting_custom_policies = procedure_config.get("delete-policies", {})
    existing_custom_policies = current_config.get("custom-policies", {})
    if existing_custom_policies or adding_custom_policies:
        # if custom policies are not defined in inventory, then we need to create custom policies ourselves
        if not existing_custom_policies:
            current_config["custom-policies"] = {}
        current_config["custom-policies"] = merge_custom_policies(existing_custom_policies,
                                                                  adding_custom_policies,
                                                                  deleting_custom_policies)

    # merge flags from procedure config and cluster config
    current_config["pod-security"] = procedure_config.get("pod-security", current_config.get("pod-security", "enabled"))
    if "oob-policies" in procedure_config:
        if "oob-policies" not in current_config:
            current_config["oob-policies"] = procedure_config["oob-policies"]
        else:
            for oob_policy in procedure_config["oob-policies"]:
                current_config["oob-policies"][oob_policy] = procedure_config["oob-policies"][oob_policy]

    return inventory_to_finalize


def merge_custom_policies(old_policies: Dict[str, List[dict]],
                          added_policies: Dict[str, List[dict]],
                          deleted_policies: Dict[str, List[dict]]) -> Dict[str, List[dict]]:
    return {
        psp_list_option: merge_policy_lists(old_policies.get(psp_list_option, []),
                                            added_policies.get(psp_list_option, []),
                                            deleted_policies.get(psp_list_option, [])),
        roles_list_option: merge_policy_lists(old_policies.get(roles_list_option, []),
                                              added_policies.get(roles_list_option, []),
                                              deleted_policies.get(roles_list_option, [])),
        bindings_list_option: merge_policy_lists(old_policies.get(bindings_list_option, []),
                                                 added_policies.get(bindings_list_option, []),
                                                 deleted_policies.get(bindings_list_option, []))
    }


def merge_policy_lists(old_list: List[dict], added_list: List[dict], deleted_list: List[dict]) -> List[dict]:
    resulting_list = added_list
    added_names_list = [item["metadata"]["name"] for item in added_list]
    deleted_names_list = [item["metadata"]["name"] for item in deleted_list]
    for old_item in old_list:
        old_item_name = old_item["metadata"]["name"]
        if old_item_name in added_names_list or old_item_name in deleted_names_list:
            # skip old item, since it was either deleted, replaced by new item, or deleted and then replaced
            continue
        # old item is nor deleted, nor updated, then we need to preserve it in resulting list
        resulting_list.append(old_item)

    return resulting_list


def install_psp_task(cluster: KubernetesCluster) -> None:
    if not is_security_enabled(cluster.inventory):
        cluster.log.debug("Pod security disabled, skipping policies installation...")
        return

    first_control_plane = cluster.nodes["control-plane"].get_first_member()

    cluster.log.debug("Installing OOB policies...")
    first_control_plane.call(manage_policies,
                      manage_type="apply",
                      manage_scope=resolve_oob_scope(cluster.inventory["rbac"]["psp"]["oob-policies"], "enabled"))

    cluster.log.debug("Installing custom policies...")
    first_control_plane.call(manage_policies,
                      manage_type="apply",
                      manage_scope=cluster.inventory["rbac"]["psp"]["custom-policies"])


def delete_custom_task(cluster: KubernetesCluster) -> None:
    if "delete-policies" not in cluster.procedure_inventory["psp"]:
        cluster.log.debug("No 'delete-policies' specified, skipping...")
        return

    cluster.log.debug("Deleting custom 'delete-policies'")
    first_control_plane = cluster.nodes["control-plane"].get_first_member()
    first_control_plane.call(manage_policies,
                      manage_type="delete",
                      manage_scope=cluster.procedure_inventory["psp"]["delete-policies"])


def add_custom_task(cluster: KubernetesCluster) -> None:
    if "add-policies" not in cluster.procedure_inventory["psp"]:
        cluster.log.debug("No 'add-policies' specified, skipping...")
        return

    cluster.log.debug("Applying custom 'add-policies'")
    first_control_plane = cluster.nodes["control-plane"].get_first_member()
    first_control_plane.call(manage_policies,
                      manage_type="apply",
                      manage_scope=cluster.procedure_inventory["psp"]["add-policies"])


def reconfigure_oob_task(cluster: KubernetesCluster) -> None:
    target_security_state = cluster.procedure_inventory["psp"].get("pod-security")
    oob_policies = cluster.procedure_inventory["psp"].get("oob-policies")

    # reconfigure OOB only if state will be changed, or OOB configuration was changed
    if not target_security_state and not oob_policies:
        cluster.log.debug("No need to reconfigure OOB policies, skipping...")
        return

    first_control_plane = cluster.nodes["control-plane"].get_first_member()

    cluster.log.debug("Deleting all OOB policies...")
    first_control_plane.call(delete_privileged_policy)
    first_control_plane.call(manage_policies, manage_type="delete", manage_scope=resolve_oob_scope(loaded_oob_policies, "all"))

    if target_security_state == "disabled":
        cluster.log.debug("Security disabled, OOB will not be recreated")
        return

    cluster.log.debug("Recreating all OOB policies...")
    policies_to_recreate = {}
    procedure_config = cluster.procedure_inventory["psp"].get("oob-policies", {})
    current_config = cluster.inventory["rbac"]["psp"]["oob-policies"]
    for policy in provided_oob_policies:
        if procedure_config.get(policy, current_config[policy]) == "enabled":
            policies_to_recreate[policy] = True
    first_control_plane.call(apply_privileged_policy)
    first_control_plane.call(manage_policies, manage_type="apply", manage_scope=resolve_oob_scope(policies_to_recreate, "all"))


def reconfigure_plugin_task(cluster: KubernetesCluster) -> None:
    target_state = cluster.procedure_inventory["psp"].get("pod-security")

    if not target_state:
        cluster.log.debug("Security plugin will not be reconfigured")
        return

    first_control_plane = cluster.nodes["control-plane"].get_first_member()

    cluster.log.debug("Updating kubeadm config map")
    final_admission_plugins_list = first_control_plane.call(update_kubeadm_configmap, target_state=target_state)

    # update api-server config on all control-planes
    cluster.log.debug("Updating kube-apiserver configs on control-planes")
    cluster.nodes["control-plane"].call(update_kubeapi_config, options_list=final_admission_plugins_list)


def restart_pods_task(cluster: KubernetesCluster) -> None:
    if cluster.context.get('initial_procedure') == 'manage_pss':
        # check if pods restart is enabled
        is_restart = cluster.procedure_inventory.get("restart-pods", False)
        if not is_restart:
            cluster.log.debug("'restart-pods' is disabled, pods won't be restarted")
            return

    first_control_plane = cluster.nodes["control-plane"].get_first_member()

    cluster.log.debug("Drain-Uncordon all nodes to restart pods")
    kube_nodes = cluster.nodes["control-plane"].include_group(cluster.nodes["worker"])
    for node in kube_nodes.get_ordered_members_list():
        first_control_plane.sudo(
            kubernetes.prepare_drain_command(cluster, node.get_node_name(), disable_eviction=False),
            hide=False)
        first_control_plane.sudo("kubectl uncordon %s" % node.get_node_name(), hide=False)

    cluster.log.debug("Restarting daemon-sets...")
    daemon_sets = ruamel.yaml.YAML().load(list(first_control_plane.sudo("kubectl get ds -A -o yaml").values())[0].stdout)
    for ds in daemon_sets["items"]:
        first_control_plane.sudo("kubectl rollout restart ds %s -n %s" % (ds["metadata"]["name"], ds["metadata"]["namespace"]))

    # we do not know to wait for, only for system pods maybe
    cluster.log.debug("Waiting for system pods...")
    kubernetes.wait_for_any_pods(cluster, first_control_plane)


def update_kubeadm_configmap_psp(first_control_plane: NodeGroup, target_state: str) -> str:
    yaml = ruamel.yaml.YAML()

    # load kubeadm config map and retrieve cluster config
    result = first_control_plane.sudo("kubectl get cm kubeadm-config -n kube-system -o yaml")
    kubeadm_cm = yaml.load(list(result.values())[0].stdout)
    cluster_config = yaml.load(kubeadm_cm["data"]["ClusterConfiguration"])

    # resolve resulting admission plugins list
    final_plugins_string = resolve_final_plugins_list(cluster_config, target_state)

    # update kubeadm config map with updated plugins list
    cluster_config["apiServer"]["extraArgs"]["enable-admission-plugins"] = final_plugins_string
    buf = io.StringIO()
    yaml.dump(cluster_config, buf)
    kubeadm_cm["data"]["ClusterConfiguration"] = buf.getvalue()

    # apply updated kubeadm config map
    buf = io.StringIO()
    yaml.dump(kubeadm_cm, buf)
    filename = uuid.uuid4().hex
    first_control_plane.put(buf, "/tmp/%s.yaml" % filename)
    first_control_plane.sudo("kubectl apply -f /tmp/%s.yaml" % filename)
    first_control_plane.sudo("rm -f /tmp/%s.yaml" % filename)

    return final_plugins_string


def update_kubeadm_configmap(first_control_plane: NodeGroup, target_state: str) -> str:
    admission_impl = first_control_plane.cluster.inventory['rbac']['admission']
    if admission_impl == "psp":
        return update_kubeadm_configmap_psp(first_control_plane, target_state)
    else:  # admission_impl == "pss":
        return update_kubeadm_configmap_pss(first_control_plane, target_state)


def update_kubeapi_config_psp(control_planes: NodeGroup, plugins_list: str) -> None:
    yaml = ruamel.yaml.YAML()

    for control_plane in control_planes.get_ordered_members_list():
        result = control_plane.sudo("cat /etc/kubernetes/manifests/kube-apiserver.yaml")

        # update kube-apiserver config with updated plugins list
        conf = yaml.load(list(result.values())[0].stdout)
        new_command = [cmd for cmd in conf["spec"]["containers"][0]["command"] if "enable-admission-plugins" not in cmd]
        new_command.append("--enable-admission-plugins=%s" % plugins_list)
        conf["spec"]["containers"][0]["command"] = new_command

        # place updated config on control-plane
        buf = io.StringIO()
        yaml.dump(conf, buf)
        control_plane.put(buf, "/etc/kubernetes/manifests/kube-apiserver.yaml", sudo=True)

    # force kube-apiserver pod restart, then wait for api to become available
        if control_planes.cluster.inventory['services']['cri']['containerRuntime'] == 'containerd':
            control_plane.call(utils.wait_command_successful,
                                                   command="crictl rm -f $(sudo crictl ps --name kube-apiserver -q)")
        else:
            control_plane.call(utils.wait_command_successful,
                                                   command="docker stop $(sudo docker ps -q -f 'name=k8s_kube-apiserver'"
                                                           " | awk '{print $1}')")
        control_plane.call(utils.wait_command_successful, command="kubectl get pod -n kube-system")


def update_kubeapi_config(control_planes: NodeGroup, options_list: str) -> None:
    admission_impl = control_planes.cluster.inventory['rbac']['admission']
    if admission_impl == "psp":
        update_kubeapi_config_psp(control_planes, options_list)
    elif admission_impl == "pss":
        update_kubeapi_config_pss(control_planes, options_list)


def is_security_enabled(inventory: dict) -> bool:
    admission_impl = inventory['rbac']['admission']
    target_state = "disabled"
    if admission_impl == "psp":
        target_state = inventory["rbac"]["psp"]["pod-security"]
    elif admission_impl == "pss":
        target_state = inventory["rbac"]["pss"]["pod-security"]

    return target_state == "enabled"


def apply_privileged_policy(group: NodeGroup) -> RunnersGroupResult:
    return manage_privileged_from_file(group, privileged_policy_filename, "apply")


def delete_privileged_policy(group: NodeGroup) -> RunnersGroupResult:
    return manage_privileged_from_file(group, privileged_policy_filename, "delete")


def apply_admission(group: NodeGroup) -> None:
    admission_impl = group.cluster.inventory['rbac']['admission']
    if is_security_enabled(group.cluster.inventory):
        if admission_impl == "psp":
            group.cluster.log.debug("Setting up privileged psp...")
            apply_privileged_policy(group)
        elif admission_impl == "pss":
            group.cluster.log.debug("Setting up default pss...")
            apply_default_pss(group.cluster)


def apply_default_pss(cluster: KubernetesCluster) -> None:
    if cluster.context.get('initial_procedure') == 'manage_pss':
        procedure_config = cluster.procedure_inventory["pss"]
        current_config = cluster.inventory["rbac"]["pss"]
        if procedure_config["pod-security"] == "enabled" and current_config["pod-security"] == "enabled":
            manage_pss(cluster, "apply")
        elif procedure_config["pod-security"] == "enabled" and current_config["pod-security"] == "disabled":
            manage_pss(cluster, "install")
    else:
        manage_pss(cluster, "init")


def delete_default_pss(cluster: KubernetesCluster) -> None:
    procedure_config = cluster.procedure_inventory["pss"]
    current_config = cluster.inventory["rbac"]["pss"]
    if procedure_config["pod-security"] == "disabled" and current_config["pod-security"] == "enabled":
        return manage_pss(cluster, "delete")


def manage_privileged_from_file(group: NodeGroup, filename: str, manage_type: str) -> RunnersGroupResult:
    if manage_type not in ["apply", "delete"]:
        raise Exception("unexpected manage type for privileged policy")
    privileged_policy = utils.read_internal(os.path.join(policies_file_path, filename))
    remote_path = tmp_filepath_pattern % filename
    group.put(io.StringIO(privileged_policy), remote_path, backup=True, sudo=True)

    return group.sudo("kubectl %s -f %s" % (manage_type, remote_path), warn=True)


def resolve_oob_scope(oob_policies_conf: Dict[str, Any], selector: str) -> Dict[str, List[dict]]:
    result: Dict[str, List[dict]] = {
        psp_list_option: [],
        roles_list_option: [],
        bindings_list_option: []
    }

    for key, value in oob_policies_conf.items():
        if value == selector or selector == "all":
            policy = loaded_oob_policies[key]
            if "psp" in policy:
                result[psp_list_option].append(policy["psp"])
            if "role" in policy:
                result[roles_list_option].append(policy["role"])
            if "binding" in policy:
                result[bindings_list_option].append(policy["binding"])

    return result


def load_oob_policies_files() -> Dict[str, dict]:
    oob_policies = {}
    for oob_name in provided_oob_policies:
        local_path = os.path.join(policies_file_path, "%s.yaml" % oob_name)
        with utils.open_internal(local_path) as stream:
            oob_policies[oob_name] = yaml.safe_load(stream)

    return oob_policies


def manage_policies(group: NodeGroup, manage_type: str,
                    manage_scope: Dict[str, List[dict]]) -> Optional[RunnersGroupResult]:
    psp_to_manage = manage_scope.get(psp_list_option, None)
    roles_to_manage = manage_scope.get(roles_list_option, None)
    bindings_to_manage = manage_scope.get(bindings_list_option, None)

    if not psp_to_manage and not roles_to_manage and not bindings_to_manage:
        group.cluster.log.verbose("No policies to %s" % manage_type)
        return None

    template = collect_policies_template(psp_to_manage, roles_to_manage, bindings_to_manage)
    filename = uuid.uuid4().hex
    remote_path = tmp_filepath_pattern % filename
    group.put(io.StringIO(template), remote_path, backup=True, sudo=True)
    result = group.sudo("kubectl %s -f %s" % (manage_type, remote_path), warn=True)
    group.sudo("rm -f %s" % remote_path)
    return result


def collect_policies_template(psp_list: Optional[List[dict]],
                              roles_list: Optional[List[dict]],
                              bindings_list: Optional[List[dict]]) -> str:
    yaml = ruamel.yaml.YAML()

    buf = io.StringIO()
    if psp_list:
        for psp in psp_list:
            yaml.dump(psp, buf)
            buf.write("\n---\n")
    if roles_list:
        for role in roles_list:
            yaml.dump(role, buf)
            buf.write("\n---\n")
    if bindings_list:
        for binding in bindings_list:
            yaml.dump(binding, buf)
            buf.write("\n---\n")
    return buf.getvalue()


def resolve_final_plugins_list(cluster_config: dict, target_state: str) -> str:
    if "enable-admission-plugins" not in cluster_config["apiServer"]["extraArgs"]:
        if target_state == "enabled":
            return "PodSecurityPolicy"
        else:
            return ""
    else:
        current_plugins = cluster_config["apiServer"]["extraArgs"]["enable-admission-plugins"]
        if "PodSecurityPolicy" not in current_plugins:
            if target_state == "enabled":
                resulting_list = "%s,%s" % (current_plugins, "PodSecurityPolicy")
            else:
                resulting_list = current_plugins
        elif target_state == "disabled":
            resulting_list = current_plugins.replace("PodSecurityPolicy", "")
        else:
            resulting_list = current_plugins

        return resulting_list.replace(",,", ",").strip(",")


def install(cluster: KubernetesCluster) -> None:
    admission_impl = cluster.inventory['rbac']['admission']
    if admission_impl == "psp":
        install_psp_task(cluster)


def manage_pss_enrichment(inventory: dict, cluster: KubernetesCluster) -> dict:
    if cluster.context.get('initial_procedure') != 'manage_pss':
        return inventory

    procedure_config = cluster.procedure_inventory["pss"]
    minor_version = int(cluster.inventory["services"]["kubeadm"]["kubernetesVersion"].split('.')[1])
        
    if not is_security_enabled(inventory) and procedure_config["pod-security"] == "disabled":
        raise Exception("both 'pod-security' in procedure config and current config are 'disabled'. There is nothing to change")

    # check flags, profiles; enrich inventory
    if minor_version < 23:
        raise Exception("PSS is not supported properly in Kubernetes version before v1.23")
    if "defaults" in procedure_config:
        for item in procedure_config["defaults"]:
            if item.endswith("version"):
                verify_version(item, procedure_config["defaults"][item], minor_version)
            inventory["rbac"]["pss"]["defaults"][item] = procedure_config["defaults"][item]
    if "exemptions" in procedure_config:
        default_merger.merge(inventory["rbac"]["pss"]["exemptions"], procedure_config["exemptions"])
    if "namespaces" in procedure_config:
        for namespace_item in procedure_config["namespaces"]:
            # check if the namespace has its own profiles
            if isinstance(namespace_item, dict):
                namespace = list(namespace_item.keys())[0]
                for item in list(namespace_item[namespace]):
                    if namespace_item[namespace][item]:
                        if item.endswith("version"):
                            verify_version(item, namespace_item[namespace][item], minor_version)
    if "namespaces_defaults" in procedure_config:
        for item in procedure_config["namespaces_defaults"]:
            if item.endswith("version"):
                verify_version(item, procedure_config["namespaces_defaults"][item], minor_version)

    return inventory


def enrich_default_admission(inventory: dict, _: KubernetesCluster) -> dict:
    minor_version = int(inventory["services"]["kubeadm"]["kubernetesVersion"].split('.')[1])
    if not inventory["rbac"].get("admission"):
        inventory["rbac"]["admission"] = "psp" if minor_version < 25 else "pss"
    return inventory


def manage_enrichment(inventory: dict, cluster: KubernetesCluster) -> dict:
    admission_impl = inventory['rbac']['admission']
    if admission_impl == "psp":
        return manage_psp_enrichment(inventory, cluster)
    elif admission_impl == "pss":
        return manage_pss_enrichment(inventory, cluster)

    return inventory


def manage_pss(cluster: KubernetesCluster, manage_type: str) -> None:
    first_control_plane = cluster.nodes["control-plane"].get_first_member()
    control_planes = cluster.nodes["control-plane"]
    # 'apply' - change options in admission.yaml, PSS is enabled
    if manage_type == "apply":
        # set labels for predifined plugins namespaces and namespaces defined in procedure config
        label_namespace_pss(cluster, manage_type)
        # copy admission config on control-planes
        copy_pss(control_planes)
        for control_plane in control_planes.get_ordered_members_list():
            # force kube-apiserver pod restart, then wait for api to become available
            if control_plane.cluster.inventory['services']['cri']['containerRuntime'] == 'containerd':
                control_plane.call(utils.wait_command_successful, command="crictl rm -f "
                                                       "$(sudo crictl ps --name kube-apiserver -q)")
            else:
                control_plane.call(utils.wait_command_successful, command="docker stop "
                                                       "$(sudo docker ps -f 'name=k8s_kube-apiserver'"
                                                       " | awk '{print $1}')")
            control_plane.call(utils.wait_command_successful, command="kubectl get pod -n kube-system")
    # 'install' - enable PSS
    elif manage_type == "install":
        # set labels for predifined plugins namespaces and namespaces defined in procedure config
        label_namespace_pss(cluster, manage_type)
        # copy admission config on control-planes
        copy_pss(cluster.nodes["control-plane"])

        cluster.log.debug("Updating kubeadm config map")
        final_features_list = first_control_plane.call(update_kubeadm_configmap_pss, target_state="enabled")

        # update api-server config on all control-planes
        cluster.log.debug("Updating kube-apiserver configs on control-planes")
        cluster.nodes["control-plane"].call(update_kubeapi_config_pss, features_list=final_features_list)
    # 'init' make changes during init Kubernetes cluster
    elif manage_type == "init":
        cluster.log.debug("Updating kubeadm config map")
        first_control_plane.call(update_kubeadm_configmap_pss, target_state="enabled")
    # 'delete' - disable PSS
    elif manage_type == "delete":
        # set labels for predifined plugins namespaces and namespaces defined in procedure config
        label_namespace_pss(cluster, manage_type)

        final_features_list = first_control_plane.call(update_kubeadm_configmap, target_state="disabled")

        # update api-server config on all control-planes
        cluster.log.debug("Updating kube-apiserver configs on control-planes")
        cluster.nodes["control-plane"].call(update_kubeapi_config_pss, features_list=final_features_list)

        # erase PSS admission config 
        cluster.log.debug("Erase admission configuration... %s" % admission_path)
        group = cluster.nodes["control-plane"]
        group.sudo("rm -f %s" % admission_path, warn=True)


def update_kubeapi_config_pss(control_planes: NodeGroup, features_list: str) -> None:
    yaml = ruamel.yaml.YAML()

    for control_plane in control_planes.get_ordered_members_list():
        result = control_plane.sudo("cat /etc/kubernetes/manifests/kube-apiserver.yaml")
        if control_plane.cluster.context['initial_procedure'] == 'upgrade':
            minor_version = int(control_plane.cluster.context['upgrade_version'].split('.')[1])
        else:
            minor_version = int(control_plane.cluster.inventory['services']['kubeadm']['kubernetesVersion'].split('.')[1])
        # update kube-apiserver config with updated features list or delete '--feature-gates' and '--admission-control-config-file'
        conf = yaml.load(list(result.values())[0].stdout)
        new_command = [cmd for cmd in conf["spec"]["containers"][0]["command"]]
        if len(features_list) != 0:
            if minor_version <= 27:
                if 'PodSecurity=true' in features_list:
                    new_command.append("--admission-control-config-file=%s" % admission_path)
                else:
                    new_command.append("--admission-control-config-file=''")
                new_command.append("--feature-gates=%s" % features_list)
            else:
                new_command.append("--admission-control-config-file=%s" % admission_path)
                if control_plane.cluster.context['initial_procedure'] == 'upgrade':
                    new_command.remove("--feature-gates=PodSecurity=true")
        else:
            for item in conf["spec"]["containers"][0]["command"]:
                if item.startswith("--"):
                    key = item.split('=')[0]
                    value = item[len(key)+1:]
                    if key in ["--feature-gates", "--admission-control-config-file"]:
                        del_option = "%s=%s" % (key, value)
                        new_command.remove(del_option)

        conf["spec"]["containers"][0]["command"] = new_command

        # place updated config on control-plane
        buf = io.StringIO()
        yaml.dump(conf, buf)
        control_plane.put(buf, "/etc/kubernetes/manifests/kube-apiserver.yaml", sudo=True)

        # force kube-apiserver pod restart, then wait for api to become available
        if control_plane.cluster.inventory['services']['cri']['containerRuntime'] == 'containerd':
            control_plane.call(utils.wait_command_successful, command="crictl rm -f "
                                                       "$(sudo crictl ps --name kube-apiserver -q)")
        else:
            control_plane.call(utils.wait_command_successful, command="docker stop "
                                                       "$(sudo docker ps -f 'name=k8s_kube-apiserver'"
                                                       " | awk '{print $1}')")
        control_plane.call(utils.wait_command_successful, command="kubectl get pod -n kube-system")


def update_kubeadm_configmap_pss(first_control_plane: NodeGroup, target_state: str) -> str:
    yaml = ruamel.yaml.YAML()

    final_feature_list = ""

    # load kubeadm config map and retrieve cluster config
    result = first_control_plane.sudo("kubectl get cm kubeadm-config -n kube-system -o yaml")
    kubeadm_cm = yaml.load(list(result.values())[0].stdout)
    cluster_config = yaml.load(kubeadm_cm["data"]["ClusterConfiguration"])
    if first_control_plane.cluster.context['initial_procedure'] == 'upgrade':
        minor_version = int(first_control_plane.cluster.context['upgrade_version'].split('.')[1])
    else:
        minor_version = int(cluster_config['kubernetesVersion'].split('.')[1])

    # update kubeadm config map with feature list
    if target_state == "enabled":
        if minor_version <= 27:
            if "feature-gates" in cluster_config["apiServer"]["extraArgs"]:
                enabled_admissions = cluster_config["apiServer"]["extraArgs"]["feature-gates"]
                if 'PodSecurity=true' not in enabled_admissions:
                    enabled_admissions = "%s,PodSecurity=true" % enabled_admissions
                    cluster_config["apiServer"]["extraArgs"]["feature-gates"] = enabled_admissions
                    cluster_config["apiServer"]["extraArgs"]["admission-control-config-file"] = admission_path
                    final_feature_list = enabled_admissions
                else:
                    cluster_config["apiServer"]["extraArgs"]["admission-control-config-file"] = admission_path
                    final_feature_list = enabled_admissions
            else:
                cluster_config["apiServer"]["extraArgs"]["feature-gates"] = "PodSecurity=true"
                cluster_config["apiServer"]["extraArgs"]["admission-control-config-file"] = admission_path
                final_feature_list = "PodSecurity=true"
        else:
            cluster_config["apiServer"]["extraArgs"]["admission-control-config-file"] = admission_path
            if first_control_plane.cluster.context['initial_procedure'] == 'upgrade':
                del cluster_config["apiServer"]["extraArgs"]["feature-gates"]
            final_feature_list = "PodSecurity deprecated in %s" % cluster_config['kubernetesVersion']
    elif target_state == "disabled":
        if minor_version <= 27:
            feature_list = cluster_config["apiServer"]["extraArgs"]["feature-gates"].replace("PodSecurity=true", "")
            final_feature_list = feature_list.replace(",,", ",")
            if len(final_feature_list) == 0:
                del cluster_config["apiServer"]["extraArgs"]["feature-gates"]
                del cluster_config["apiServer"]["extraArgs"]["admission-control-config-file"]
            else:
                cluster_config["apiServer"]["extraArgs"]["feature-gates"] = final_feature_list
                del cluster_config["apiServer"]["extraArgs"]["admission-control-config-file"]
        else:
            if cluster_config["apiServer"]["extraArgs"].get("feature-gates"):
                if len(cluster_config["apiServer"]["extraArgs"]["feature-gates"]) == 0:
                    del cluster_config["apiServer"]["extraArgs"]["feature-gates"]
                    del cluster_config["apiServer"]["extraArgs"]["admission-control-config-file"]
                else:
                    del cluster_config["apiServer"]["extraArgs"]["admission-control-config-file"]
            else:
                del cluster_config["apiServer"]["extraArgs"]["admission-control-config-file"]

    buf = io.StringIO()
    yaml.dump(cluster_config, buf)
    kubeadm_cm["data"]["ClusterConfiguration"] = buf.getvalue()

    # apply updated kubeadm config map
    buf = io.StringIO()
    yaml.dump(kubeadm_cm, buf)
    filename = uuid.uuid4().hex
    first_control_plane.put(buf, "/tmp/%s.yaml" % filename)
    first_control_plane.sudo("kubectl apply -f /tmp/%s.yaml" % filename)
    first_control_plane.sudo("rm -f /tmp/%s.yaml" % filename)

    return final_feature_list


def finalize_inventory(cluster: KubernetesCluster, inventory_to_finalize: dict) -> dict:
    admission_impl = cluster.inventory['rbac']['admission']

    if admission_impl == "psp":
        return finalize_inventory_psp(cluster, inventory_to_finalize)
    elif admission_impl == "pss":
        return finalize_inventory_pss(cluster, inventory_to_finalize)

    return inventory_to_finalize


def finalize_inventory_pss(cluster: KubernetesCluster, inventory_to_finalize: dict) -> dict:
    if cluster.context.get('initial_procedure') != 'manage_pss':
        return inventory_to_finalize
    procedure_config = cluster.procedure_inventory["pss"]

    current_config = inventory_to_finalize.setdefault("rbac", {}).setdefault("pss", {})

    # merge flags from procedure config and cluster config
    current_config["pod-security"] = procedure_config.get("pod-security", current_config.get("pod-security", "enabled"))
    if "defaults" in procedure_config:
        default_merger.merge(current_config.setdefault("defaults", {}), procedure_config["defaults"])
    if "exemptions" in procedure_config:
        default_merger.merge(current_config.setdefault("exemptions", {}), procedure_config["exemptions"])

    return inventory_to_finalize


# update PSP/PSS fields in the inventory dumped to cluster_finalized.yaml
def update_finalized_inventory(cluster: KubernetesCluster, inventory_to_finalize: dict) -> dict:
    if cluster.context.get('initial_procedure') == 'manage_pss':
        current_config = inventory_to_finalize.setdefault("rbac", {}).setdefault("pss", {})
        current_config["pod-security"] = cluster.procedure_inventory["pss"].get("pod-security", current_config.get("pod-security", "enabled"))
    elif cluster.context.get('initial_procedure') == 'manage_psp':
        current_config = inventory_to_finalize.setdefault("rbac", {}).setdefault("psp", {})
        current_config["pod-security"] = cluster.procedure_inventory["psp"].get("pod-security", current_config.get("pod-security", "enabled"))
    # remove PSP section from cluster_finalyzed.yaml  
    minor_version = int(inventory_to_finalize["services"]["kubeadm"]["kubernetesVersion"].split('.')[1])
    if minor_version > 24:
        del inventory_to_finalize["rbac"]["psp"]

    return inventory_to_finalize


def copy_pss(group: NodeGroup) -> Optional[RunnersGroupResult]:
    if  group.cluster.inventory['rbac']['admission'] != "pss":
        return None
    if group.cluster.context.get('initial_procedure') == 'manage_pss':
        if not is_security_enabled(group.cluster.inventory) and \
                group.cluster.procedure_inventory["pss"]["pod-security"] != "enabled":
            group.cluster.log.debug("Pod security disabled, skipping pod admission installation...")
            return None
    if group.cluster.context.get('initial_procedure') == 'install':
        if not is_security_enabled(group.cluster.inventory):
            group.cluster.log.debug("Pod security disabled, skipping pod admission installation...")
            return None

    defaults = group.cluster.inventory["rbac"]["pss"]["defaults"]
    exemptions = group.cluster.inventory["rbac"]["pss"]["exemptions"]
    # create admission config from template and cluster.yaml
    admission_config = Template(utils.read_internal(admission_template))\
                       .render(defaults=defaults,exemptions=exemptions)

    # put admission config on every control-planes
    filename = uuid.uuid4().hex
    remote_path = tmp_filepath_pattern % filename
    group.cluster.log.debug("Copy admission config: %s, %s" % (remote_path, admission_path))
    group.put(io.StringIO(admission_config), remote_path, backup=True, sudo=True)
    group.sudo("mkdir -p %s" % admission_dir, warn=True)
    result = group.sudo("cp %s %s" % (remote_path, admission_path), warn=True)
    group.sudo("rm -f %s" % remote_path)

    return result


def label_namespace_pss(cluster: KubernetesCluster, manage_type: str) -> None:
    first_control_plane = cluster.nodes["control-plane"].get_first_member()
    # set/delete labels on predifined plugins namsespaces
    for plugin in cluster.inventory["plugins"]:
        is_install = cluster.inventory["plugins"][plugin].get("install")
        if manage_type in ["apply", "install"]:
            if is_install and plugin in privileged_plugins.keys():
                # set label 'pod-security.kubernetes.io/enforce: privileged' for local provisioner and ingress namespaces
                cluster.log.debug(f"Set PSS labels on namespace {privileged_plugins[plugin]}")
                for mode in valid_modes:
                    first_control_plane.sudo(f"kubectl label ns {privileged_plugins[plugin]} "
                                             f"pod-security.kubernetes.io/{mode}=privileged --overwrite")
                    first_control_plane.sudo(f"kubectl label ns {privileged_plugins[plugin]} "
                                             f"pod-security.kubernetes.io/{mode}-version=latest --overwrite")
            elif is_install and plugin in baseline_plugins.keys():
                # set label 'pod-security.kubernetes.io/enforce: baseline' for kubernetes dashboard
                cluster.log.debug(f"Set PSS labels on namespace {baseline_plugins[plugin]}")
                for mode in valid_modes:
                    first_control_plane.sudo(f"kubectl label ns {baseline_plugins[plugin]} "
                                             f"pod-security.kubernetes.io/{mode}=baseline --overwrite")
                    first_control_plane.sudo(f"kubectl label ns {baseline_plugins[plugin]} "
                                             f"pod-security.kubernetes.io/{mode}-version=latest --overwrite")
        elif manage_type == "delete":
            if is_install and plugin in privileged_plugins.keys():
                # delete label 'pod-security.kubernetes.io/enforce: privileged' for local provisioner and ingress namespaces
                cluster.log.debug(f"Delete PSS labels from namespace {privileged_plugins[plugin]}")
                for mode in valid_modes:
                    first_control_plane.sudo(f"kubectl label ns {privileged_plugins[plugin]} "
                                             f"pod-security.kubernetes.io/{mode}- || true")
                    first_control_plane.sudo(f"kubectl label ns {privileged_plugins[plugin]} "
                                             f"pod-security.kubernetes.io/{mode}-version- || true")
            elif is_install and plugin in baseline_plugins.keys():
                # delete 'pod-security.kubernetes.io/enforce: baseline' for kubernetes dashboard
                cluster.log.debug(f"Delete PSS labels from namespace {baseline_plugins[plugin]}")
                for mode in valid_modes:
                    first_control_plane.sudo(f"kubectl label ns {baseline_plugins[plugin]} "
                                             f"pod-security.kubernetes.io/{mode}- || true")
                    first_control_plane.sudo(f"kubectl label ns {baseline_plugins[plugin]} "
                                             f"pod-security.kubernetes.io/{mode}-version- || true")

    procedure_config = cluster.procedure_inventory["pss"]
    namespaces = procedure_config.get("namespaces")
    # get the list of namespaces that should be labeled then set/delete labels
    if namespaces:
        default_modes = {}
        # check if procedure config has default values for labels
        namespaces_defaults = procedure_config.get("namespaces_defaults")
        if namespaces_defaults:
            for default_mode in namespaces_defaults:
                 default_modes[default_mode] = namespaces_defaults[default_mode]
        for namespace in namespaces:
            # define name of namespace
            if isinstance(namespace, dict):
                ns_name = list(namespace.keys())[0]
            else:
                ns_name = namespace
            if manage_type in ["apply", "install"]:
                if default_modes:
                    # set labels that are set in default section
                    cluster.log.debug(f"Set PSS labels on {ns_name} namespace from defaults")
                    for mode in default_modes:
                        first_control_plane.sudo(f"kubectl label ns {ns_name} "
                                f"pod-security.kubernetes.io/{mode}={default_modes[mode]} --overwrite")
                if isinstance(namespace, dict):
                    # set labels that are set in namespaces section
                    cluster.log.debug(f"Set PSS labels on {ns_name} namespace")
                    for item in list(namespace[ns_name]):
                        first_control_plane.sudo(f"kubectl label ns {ns_name} " 
                                    f"pod-security.kubernetes.io/{item}={namespace[ns_name][item]} --overwrite")
            elif manage_type == "delete":
                # delete labels that are set in default section
                if default_modes:
                    cluster.log.debug(f"Delete PSS labels on {ns_name} namespace from defaults")
                    for mode in default_modes:
                        first_control_plane.sudo(f"kubectl label ns {ns_name} pod-security.kubernetes.io/{mode}-")
                # delete labels that are set in namespaces section
                cluster.log.debug(f"Delete PSS labels on {ns_name} namespace")
                if isinstance(namespace, dict):
                    for item in list(namespace[ns_name]):
                        first_control_plane.sudo(f"kubectl label ns {ns_name} "
                                    f"pod-security.kubernetes.io/{item}-")


def check_inventory(cluster: KubernetesCluster) -> None:
    # check if 'admission' option in cluster.yaml and procedure.yaml are inconsistent 
    if cluster.context.get('initial_procedure') == 'manage_pss' and cluster.inventory["rbac"]["admission"] != "pss" or \
        cluster.context.get('initial_procedure') == 'manage_psp' and cluster.inventory["rbac"]["admission"] != "psp":
        raise Exception("Procedure config and cluster config are inconsistent. Please check 'admission' option")



# Copyright 2021-2023 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from textwrap import dedent

from kubemarine.core.action import Action
from kubemarine.core.patch import RegularPatch
from kubemarine.core.resources import DynamicResources
from kubemarine.thirdparties import install_thirdparty
from kubemarine.cri.containerd import configure_ctr_flags


class TheAction(Action):
    def __init__(self) -> None:
        super().__init__("Replace podman to ctr")

    def run(self, res: DynamicResources) -> None:
        cluster = res.cluster()

        if cluster.inventory["services"]["cri"]['containerRuntime'] != 'containerd':
            cluster.log.info("Used not containerd cri: migration to ctr is not needed")
            return
        cluster.log.info("Load ctr options file")
        configure_ctr_flags(cluster.make_group_from_roles(["control-plane", "worker"]))

        cluster.log.info("Update /usr/bin/etcdctl thirdparty")
        install_thirdparty(cluster.nodes['all'], '/usr/bin/etcdctl')


class ReplacePodmanToCtr(RegularPatch):
    def __init__(self) -> None:
        super().__init__("replace_podman_to_ctr")

    @property
    def action(self) -> Action:
        return TheAction()

    @property
    def description(self) -> str:
        return dedent(
            f"""\
            This patch does following changes for containerd cri:
            1. Load ctr options file to control-plane and worker nodes
            2. Reinstall /usr/bin/etcdctl thirdparty to use ctr tool inside
            """.rstrip()
        )



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import copy

from typing import Any, OrderedDict, List

from kubemarine import kubernetes, packages
from kubemarine.core import flow, utils
from kubemarine.core.action import Action
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.resources import DynamicResources
from kubemarine.procedures import install


def deploy_kubernetes_join(cluster: KubernetesCluster) -> None:

    group = cluster.make_group_from_roles(['control-plane', 'worker']).get_new_nodes()

    if group.is_empty():
        cluster.log.debug("No kubernetes nodes to perform")
        return

    cluster.nodes['control-plane'].get_new_nodes().call(kubernetes.join_new_control_plane)

    if "worker" in cluster.nodes:
        cluster.nodes['worker'].get_new_nodes().exclude_group(cluster.nodes['control-plane']) \
            .call(kubernetes.init_workers)

    group.call_batch([
        kubernetes.apply_labels,
        kubernetes.apply_taints
    ])

    cluster.log.debug("Waiting for new kubernetes nodes...")
    kubernetes.wait_for_nodes(group)
    kubernetes.schedule_running_nodes_report(cluster)


def add_node_finalize_inventory(cluster: KubernetesCluster, inventory_to_finalize: dict) -> dict:
    if cluster.context.get('initial_procedure') != 'add_node':
        return inventory_to_finalize

    new_nodes = cluster.nodes['all'].get_new_nodes()

    # add nodes to inventory if they in new nodes
    for new_node in new_nodes.get_ordered_members_list():
        new_node_name = new_node.get_node_name()
        new_node_found = False
        for i, node in enumerate(inventory_to_finalize['nodes']):
            if node['name'] == new_node_name:
                # new node already presented in final inventory - ok, just remove label
                if 'add_node' in inventory_to_finalize['nodes'][i]['roles']:
                    inventory_to_finalize['nodes'][i]['roles'].remove('add_node')
                new_node_found = True
                break

        # new node is not presented in final inventory - let's add it original config
        if not new_node_found:
            node_config = None

            # search for new node config in procedure inventory
            if cluster.procedure_inventory.get('nodes', {}):
                for node_from_procedure in cluster.procedure_inventory['nodes']:
                    if node_from_procedure['name'] == new_node_name:
                        node_config = node_from_procedure
                        break
            # maybe new nodes from other places?

            if node_config is None:
                raise Exception('Not possible to find new node config for final inventory')
            inventory_to_finalize["nodes"].append(node_config)

    # maybe merge vrrp ips only when adding?
    if "vrrp_ips" in cluster.procedure_inventory:
        utils.merge_vrrp_ips(cluster.procedure_inventory, inventory_to_finalize)

    return inventory_to_finalize


def cache_installed_packages(cluster: KubernetesCluster) -> None:
    """
    Task which is used to collect already installed packages versions on already existing nodes.
    It is called first during "add_node" procedure,
    so that new nodes install exactly the same packages as on other already existing nodes.
    """
    packages.cache_package_versions(cluster, cluster.inventory, by_initial_nodes=True)


tasks: OrderedDict[str, Any] = collections.OrderedDict(copy.deepcopy(install.tasks))
del tasks["deploy"]["plugins"]
del tasks["deploy"]["accounts"]
tasks["deploy"]["kubernetes"]["init"] = deploy_kubernetes_join
tasks["cache_packages"] = cache_installed_packages
tasks.move_to_end("cache_packages", last=False)


class AddNodeAction(Action):
    def __init__(self) -> None:
        super().__init__('add node', recreate_inventory=True)

    def run(self, res: DynamicResources) -> None:
        flow.run_tasks(res, tasks, cumulative_points=install.cumulative_points)
        res.make_final_inventory()


def main(cli_arguments: List[str] = None) -> None:

    cli_help = '''
    Script for adding node to Kubernetes cluster.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, tasks=tasks)
    context = flow.create_context(parser, cli_arguments, procedure='add_node')

    flow.ActionsFlow([AddNodeAction()]).run_flow(context)


if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
from typing import Callable, Optional, List, IO, Tuple, cast

import ruamel.yaml
import os
from abc import ABC, abstractmethod

from ordered_set import OrderedSet

from kubemarine import plugins
from kubemarine.core import utils, log
from kubemarine.core.cluster import KubernetesCluster

ERROR_MANIFEST_NOT_FOUND = "Cannot find original manifest %s for '%s' plugin"


def get_default_manifest_path(plugin_name: str, version: str) -> str:
    resource = f"plugins/yaml/{plugin_name}-{version}-original.yaml"
    return utils.get_internal_resource_path(resource)


class Manifest:
    def __init__(self, stream: IO) -> None:
        self._patched = OrderedSet[str]()
        self._excluded = OrderedSet[str]()
        self._included = OrderedSet[str]()
        self._obj_list = self._load(stream)

    def obj_key(self, obj: dict) -> str:
        return f"{obj['kind']}_{obj['metadata']['name']}"

    def all_obj_keys(self) -> List[str]:
        return [self.obj_key(obj) for obj in self._obj_list]

    def key_index(self, key: str) -> int:
        for i, obj in enumerate(self._obj_list):
            if self.obj_key(obj) == key:
                return i

        raise ValueError(f"{key} not found")

    def has_obj(self, key: str) -> bool:
        return any(self.obj_key(obj) == key for obj in self._obj_list)

    def get_obj(self, key: str, *, patch: bool) -> dict:
        """
        Get manifest object in YAML format for the specified key.
        By default, ensure presence of the object.
        If the object is absent, it may be a sign that we no longer able to fulfill the previous contract.
        Caller methods must explicitly allow absent object when necessary.

        :param key: 'kind' and 'name' of object
        :param patch: boolean value that tells the manifest that the searched object is going to be patched.
        :return: manifest object
        """
        for obj in self._obj_list:
            if self.obj_key(obj) == key:
                if patch:
                    self._patched.add(key)
                return obj

        raise ValueError(f"{key} not found")

    def include(self, index: int, obj: dict) -> None:
        self._obj_list.insert(index, obj)
        key = self.obj_key(obj)
        self._included.add(key)

    def exclude(self, key: str) -> None:
        del self._obj_list[self.key_index(key)]
        self._excluded.add(key)

    @property
    def patched(self) -> List[str]:
        return list(self._patched)

    @property
    def included(self) -> List[str]:
        return list(self._included)

    @property
    def excluded(self) -> List[str]:
        return list(self._excluded)

    def dump(self) -> str:
        """
        The method implements the dumping of the list of objects to the string that includes several YAMLs inside
        """
        yaml = ruamel.yaml.YAML()

        with io.StringIO() as stream:
            yaml.dump_all(self._obj_list, stream)
            result = stream.getvalue()

        return result

    def get_all_container_images(self) -> List[str]:
        images = []
        for key in self.all_obj_keys():
            obj = self.get_obj(key, patch=False)
            for spec_section in ('containers', 'initContainers'):
                containers = obj.get('spec', {}).get('template', {}).get('spec', {}).get(spec_section, [])
                for container in containers:
                    image = container['image']
                    if image not in images:
                        images.append(image)

        return images

    def _load(self, stream: IO) -> List[dict]:
        """
        The method implements the parse YAML file that includes several YAMLs inside
        :param stream: stream with manifest content that should be parsed
        :return: list of original objects to enrich in YAML format.
        """
        yaml = ruamel.yaml.YAML()
        obj_list = []
        source_yamls = yaml.load_all(stream)
        yaml_keys = set()
        for source_yaml in source_yamls:
            if source_yaml is None:
                continue
            yaml_key = self.obj_key(source_yaml)
            # check if there is no duplication
            if yaml_key in yaml_keys:
                raise Exception(
                    f"The {yaml_key} object is duplicated, please verify the original yaml")

            yaml_keys.add(yaml_key)
            obj_list.append(source_yaml)

        return obj_list


EnrichmentFunction = Callable[[Manifest], None]


class Processor(ABC):
    def __init__(self, logger: log.VerboseLogger, inventory: dict, plugin_name: str,
                 original_yaml_path: Optional[str], destination_name: Optional[str]) -> None:
        """
        :param logger: VerboseLogger instance
        :param inventory: inventory of the cluster
        :param plugin_name: name of plugin-owner
        :param original_yaml_path: path to custom manifest
        :param destination_name: custom destination manifest file name
        """
        self.log: log.VerboseLogger = logger
        self.inventory = inventory
        self.plugin_name = plugin_name
        self.manifest_path = self._get_manifest_path(original_yaml_path)
        self.destination_name = self._get_destination(destination_name)

    def get_known_objects(self) -> List[str]:
        """
        :return: list with the 'kind' and 'name' of expected objects in original manifest
        """
        return []

    @abstractmethod
    def get_enrichment_functions(self) -> List[EnrichmentFunction]:
        """
        :return: list of enrichment methods
        """
        pass

    def get_version(self) -> str:
        version: str = self.inventory['plugins'][self.plugin_name]['version']
        return version

    def include(self, manifest: Manifest, index: int, obj: dict) -> None:
        key = manifest.obj_key(obj)
        manifest.include(index, obj)
        self.log.verbose(f"The {key} has been added")

    def exclude(self, manifest: Manifest, key: str) -> None:
        manifest.exclude(key)
        self.log.verbose(f"The {key} has been excluded from result")

    def validate_inventory(self) -> None:
        """
        # Check if original YAML exists
        """
        if not os.path.isfile(self.manifest_path):
            raise Exception(ERROR_MANIFEST_NOT_FOUND % (self.manifest_path, self.plugin_name))

    def validate_original(self, manifest: Manifest) -> None:
        """
        The method implements default validations for manifest objects

        :param manifest: container to operate with manifest objects
        """
        known_objects = self.get_known_objects()

        # check if there are new objects
        for key in manifest.all_obj_keys():
            if key not in known_objects:
                self.log.verbose(f"The current version of original yaml has a new object: {key}")

        # check if known objects were excluded
        for key in known_objects:
            if not manifest.has_obj(key):
                self.log.verbose(f"The current version of original yaml does not include "
                                 f"the following object: {key}")

    def enrich(self) -> Manifest:
        """
        The method implements full processing for the plugin main manifest.
        """

        # get original YAML and parse it into list of objects
        try:
            with utils.open_utf8(self.manifest_path, 'r') as stream:
                manifest = Manifest(stream)
        except Exception as exc:
            raise Exception(f"Failed to load {self.manifest_path} for {self.plugin_name!r} plugin") from exc

        self.validate_original(manifest)

        # call enrichment functions one by one
        enrichment_functions = self.get_enrichment_functions()
        for fn in enrichment_functions:
            fn(manifest)

        self.log.verbose(f"The total number of patched objects is {len(manifest.patched)} "
                         f"the objects are the following: {manifest.patched}")
        self.log.verbose(f"The total number of added objects is {len(manifest.included)} "
                         f"the objects are the following: {manifest.included}")
        self.log.verbose(f"The total number of excluded objects is {len(manifest.excluded)} "
                         f"the objects are the following: {manifest.excluded}")

        return manifest

    def apply(self, cluster: KubernetesCluster, manifest: Manifest) -> None:
        logger = cluster.log
        enriched_manifest = manifest.dump()
        utils.dump_file(cluster, enriched_manifest, self.destination_name)

        destination = '/etc/kubernetes/%s' % self.destination_name

        # create config for plugin module
        config = {
            "source": io.StringIO(enriched_manifest),
            "destination": destination,
            "do_render": False
        }

        logger.debug(f"Uploading manifest enriched from {self.manifest_path} for {self.plugin_name!r} plugin...")
        logger.debug("\tDestination: %s" % destination)

        plugins.apply_source(cluster, config)

    def _get_manifest_path(self, custom_manifest_path: Optional[str]) -> str:
        if custom_manifest_path is not None:
            config = {"source": custom_manifest_path}
            manifest_path, _ = plugins.get_source_absolute_pattern(config)
        else:
            manifest_path = get_default_manifest_path(self.plugin_name, self.get_version())

        return manifest_path

    def _get_destination(self, custom_destination_name: Optional[str]) -> str:
        if custom_destination_name is not None:
            return custom_destination_name

        return f'{self.plugin_name}-{self.get_version()}.yaml'

    def assign_default_pss_labels(self, manifest: Manifest, key: str, profile: str) -> None:
        source_yaml = manifest.get_obj(key, patch=True)
        labels: dict = source_yaml['metadata'].setdefault('labels', {})
        labels.update({
            'pod-security.kubernetes.io/enforce': profile,
            'pod-security.kubernetes.io/enforce-version': 'latest',
            'pod-security.kubernetes.io/audit': profile,
            'pod-security.kubernetes.io/audit-version': 'latest',
            'pod-security.kubernetes.io/warn': profile,
            'pod-security.kubernetes.io/warn-version': 'latest',
        })
        self.log.verbose(f"The {key} has been patched in 'metadata.labels' with pss labels for {profile!r} profile")

    def find_container_for_patch(self, manifest: Manifest, key: str,
                                 *,
                                 container_name: str, is_init_container: bool) -> Tuple[int, dict]:
        """
        Find container according to the search criteria.

        :param manifest: container to operate with manifest objects
        :param key: 'kind' and 'name' of object
        :param container_name: name of container to assign the image in the spec
        :param is_init_container: whether to search container in 'initContainers' or in 'containers' spec.
        :return: tuple of container index within the spec and the container data.
        """
        pos, container = self._find_optional_container(
            manifest, key,
            container_name=container_name, is_init_container=is_init_container, allow_absent=False)

        return pos, cast(dict, container)

    def _find_optional_container(self, manifest: Manifest, key: str,
                                 container_name: str, is_init_container: bool, allow_absent: bool = False) \
            -> Tuple[int, Optional[dict]]:
        source_yaml = manifest.get_obj(key, patch=True)
        template_spec = source_yaml['spec']['template']['spec']
        spec_containers_section = 'initContainers' if is_init_container else 'containers'
        container_pos, container = next(((i, c) for i, c in enumerate(template_spec[spec_containers_section])
                                         if c['name'] == container_name),
                                        (-1, None))

        if container_pos == -1 and not allow_absent:
            raise ValueError(f"Container {container_name!r} is not found in {spec_containers_section!r} spec of {key}")

        return container_pos, container

    def get_target_image(self, plugin_service: Optional[str] = None, image_key: Optional[str] = 'image') -> str:
        """
        Calculates full image path from the plugin configuration in inventory.

        :param plugin_service: section of plugin that contains the desirable image_key
        :param image_key: property name by which the inventory holds the desirable image
        :return: target image to be used in a container
        """
        plugin_section = self.inventory['plugins'][self.plugin_name]
        registry = plugin_section['installation'].get('registry')
        plugin_service_section = plugin_section
        if plugin_service:
            plugin_service_section = plugin_service_section[plugin_service]
        image: str = plugin_service_section[image_key]
        if registry:
            image = f"{registry}/{image}"

        return image

    def enrich_image_for_container(self, manifest: Manifest, key: str,
                                   *,
                                   plugin_service: Optional[str] = None,
                                   container_name: str, is_init_container: bool,
                                   allow_absent: bool = False) -> None:
        """
        The method patches the image of the specified container.

        :param manifest: container to operate with manifest objects
        :param key: 'kind' and 'name' of object
        :param plugin_service: section of plugin that contains the desirable 'image'
        :param container_name: name of container to assign the image in the spec
        :param is_init_container: whether to search container in 'initContainers' or in 'containers' spec.
        :param allow_absent: if True, and if container is not found, silently do nothing.
        """
        image = self.get_target_image(plugin_service=plugin_service, image_key='image')

        spec_containers_section = 'initContainers' if is_init_container else 'containers'

        container_pos, container = self._find_optional_container(manifest, key,
            container_name=container_name, is_init_container=is_init_container, allow_absent=allow_absent)
        if container is None:
            return

        container['image'] = image
        self.log.verbose(f"The {key} has been patched in "
                         f"'spec.template.spec.{spec_containers_section}.[{container_pos}].image' with {image!r}")

    def enrich_resources_for_container(self, manifest: Manifest, key: str,
                                       *,
                                       plugin_service: Optional[str] = None,
                                       container_name: str) -> None:
        """
        The method patches the resources of the specified container.

        :param manifest: container to operate with manifest objects
        :param key: 'kind' and 'name' of object
        :param plugin_service: section of plugin that contains the desirable 'resources'
        :param container_name: name of container to assign the resources in the spec
        """
        container_pos, container = self.find_container_for_patch(manifest, key,
                                                                 container_name=container_name, is_init_container=False)

        plugin_service_section = self.inventory['plugins'][self.plugin_name]
        if plugin_service:
            plugin_service_section = plugin_service_section[plugin_service]
        container['resources'] = plugin_service_section['resources']

        self.log.verbose(f"The {key} has been patched in "
                         f"'spec.template.spec.containers.[{container_pos}].resources' with {plugin_service_section['resources']!r}")

    def enrich_node_selector(self, manifest: Manifest, key: str,
                             *,
                             plugin_service: str) -> None:
        source_yaml = manifest.get_obj(key, patch=True)
        node_selector = self.inventory['plugins'][self.plugin_name][plugin_service]['nodeSelector']
        source_yaml['spec']['template']['spec']['nodeSelector'] = node_selector
        self.log.verbose(f"The {key} has been patched in 'spec.template.spec.nodeSelector' with {node_selector!r}")

    def enrich_tolerations(self, manifest: Manifest, key: str,
                           *,
                           plugin_service: Optional[str] = None,
                           extra_tolerations: List[dict] = None,
                           override: bool = False) -> None:
        source_yaml = manifest.get_obj(key, patch=True)
        template_spec: dict = source_yaml['spec']['template']['spec']
        if override and template_spec.get('tolerations', []):
            del template_spec['tolerations']
            self.log.verbose(f"The 'tolerations' property has been removed from 'spec.template.spec' in the {key}")

        tolerations: List[dict] = []
        if extra_tolerations:
            tolerations.extend(extra_tolerations)
        plugin_service_section = self.inventory['plugins'][self.plugin_name]
        if plugin_service:
            plugin_service_section = plugin_service_section[plugin_service]
        tolerations.extend(plugin_service_section.get('tolerations', []))

        for val in tolerations:
            template_spec.setdefault('tolerations', []).append(val)
            self.log.verbose(f"The {key} has been patched in 'spec.template.spec.tolerations' with '{val}'")


PROCESSOR_PROVIDER = Callable[[log.VerboseLogger, dict, Optional[str], Optional[str]], Processor]



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import io
import os
import random
import uuid
from abc import ABC, abstractmethod
from types import FunctionType
from typing import (
    Callable, Dict, List, Union, Any, TypeVar, Mapping, Iterator, Optional, Iterable, Generic, Set, cast
)

from kubemarine.core import utils, log, errors
from kubemarine.core.connections import ConnectionPool
from kubemarine.core.executor import (
    RawExecutor, Token, GenericResult, RunnersResult, HostToResult, Callback, TokenizedResult,
)

NodeConfig = Dict[str, Any]
GroupFilter = Union[Callable[[NodeConfig], bool], NodeConfig]

_RESULT = TypeVar('_RESULT', bound=GenericResult, covariant=True)
_T = TypeVar('_T')


class GenericGroupResult(Mapping[str, _RESULT]):

    def __init__(self, cluster: object, results: Mapping[str, _RESULT]) -> None:
        self.cluster = cluster
        self._result: Dict[str, _RESULT] = dict(results)

    def __getitem__(self, host: str) -> _RESULT:
        return self._result[host]

    def __len__(self) -> int:
        return len(self._result)

    def __iter__(self) -> Iterator[str]:
        return iter(self._result)

    def get_simple_out(self) -> str:
        return self.get_simple_result().stdout

    def get_simple_result(self) -> RunnersResult:
        if len(self) != 1:
            raise NotImplementedError("Simple result can be returned only for NodeGroupResult consisted of "
                                      "exactly one node, but %s were provided." % list(self.keys()))

        res = list(self.values())[0]
        if not isinstance(res, RunnersResult):
            raise NotImplementedError("It does not make sense to return simple output for result of type %s"
                                      % type(res))

        return res

    def __str__(self) -> str:
        host_outputs = []
        for host, result in self.items():
            output = f"{host}:"

            if isinstance(result, RunnersResult):
                output += f" code={result.repr_code()}"
                repr_out = result.repr_out()
                if repr_out:
                    output += '\n\t' + repr_out.replace('\n', '\n\t')

            # The exception may be only the last in the list. We are also sure to have at least one result per node.
            if isinstance(result, Exception):
                exception = errors.wrap_kme_exception(result)
                output += '\n\t' + str(exception).replace('\n', '\n\t')

            host_outputs.append(output)

        return "\n".join(host_outputs)

    def _make_group(self, hosts: Iterable[str]) -> NodeGroup:
        return NodeGroup(hosts, self.cluster)

    def get_group(self) -> NodeGroup:
        """
        Forms and returns a new group from node results
        :return: NodeGroup
        """
        return self._make_group(self.keys())

    def is_any_has_code(self, code: int) -> bool:
        """
        Returns true if some group result has an exit code equal to the given one. Exceptions and other objects in
        results will be ignored.
        :param code: The code with which the result codes will be compared
        :return: Boolean
        """
        for result in self.values():
            if isinstance(result, RunnersResult) and result.exited == code:
                return True
        return False

    def is_any_failed(self) -> bool:
        """
        Returns true if at least one result in the group finished with non-zero code
        :return: Boolean
        """
        return len(self.get_failed_hosts_list()) > 0

    def get_failed_hosts_list(self) -> List[str]:
        """
        Returns a list of hosts whose result finished with non-zero code
        :return: List with hosts
        """
        failed_hosts: List[str] = []
        for host, result in self.items():
            if isinstance(result, RunnersResult) and result.failed:
                failed_hosts.append(host)
        return failed_hosts

    def get_failed_nodes_group(self) -> NodeGroup:
        """
        Forms and returns new NodeGroup of nodes that either exited with an exception, or the exit code is equals 1
        :return: NodeGroup:
        """
        nodes_list = self.get_failed_hosts_list()
        return self._make_group(nodes_list)

    def get_hosts_list_where_value_in_stdout(self, value: str) -> List[str]:
        """
        Returns a list of hosts that contains the given string value in results stderr.
        :param value: The string value to be found in the nodes results stdout.
        :return: List with hosts
        """
        hosts_with_stderr_value: List[str] = []
        for host, result in self.items():
            if isinstance(result, RunnersResult) and value in result.stdout:
                hosts_with_stderr_value.append(host)
        return hosts_with_stderr_value

    def get_hosts_list_where_value_in_stderr(self, value: str) -> List[str]:
        """
        Returns a list of hosts that contains the given string value in results stderr.
        :param value: The string value to be found in the nodes results stderr.
        :return: List with hosts
        """
        hosts_with_stderr_value: List[str] = []
        for host, result in self.items():
            if isinstance(result, RunnersResult) and value in result.stderr:
                hosts_with_stderr_value.append(host)
        return hosts_with_stderr_value

    def get_nodes_group_where_value_in_stdout(self, value: str) -> NodeGroup:
        """
        Forms and returns new NodeGroup of nodes that contains the given string value in results stdout.
        :param value: The string value to be found in the nodes results stdout.
        :return: NodeGroup
        """
        nodes_list = self.get_hosts_list_where_value_in_stdout(value)
        return self._make_group(nodes_list)

    def get_nodes_group_where_value_in_stderr(self, value: str) -> NodeGroup:
        """
        Forms and returns new NodeGroup of nodes that contains the given string value in results stderr.
        :param value: The string value to be found in the nodes results stderr.
        :return: NodeGroup
        """
        nodes_list = self.get_hosts_list_where_value_in_stderr(value)
        return self._make_group(nodes_list)

    def stdout_contains(self, value: str) -> bool:
        """
        Checks for the presence of the given string in all results stdout.
        :param value: The string value to be found in the nodes results stdout.
        :return: true if string presented
        """
        return len(self.get_hosts_list_where_value_in_stdout(value)) > 0

    def stderr_contains(self, value: str) -> bool:
        """
        Checks for the presence of the given string in all results stderr.
        :param value: The string value to be found in the nodes results stderr.
        :return: true if string presented
        """
        return len(self.get_hosts_list_where_value_in_stderr(value)) > 0

    def __eq__(self, other: object) -> bool:
        if self is other:
            return True

        if not isinstance(other, GenericGroupResult):
            return False

        if len(self) != len(other):
            return False

        for host, result in self.items():
            compared_result = other.get(host)
            if compared_result is None:
                return False

            if not isinstance(result, RunnersResult) or not isinstance(compared_result, RunnersResult):
                raise NotImplementedError('Currently only instances of RunnersResult can be compared')

            if result != compared_result:
                return False

        return True

    def __ne__(self, other: object) -> bool:
        return not self == other


class NodeGroupResult(GenericGroupResult[GenericResult]):
    pass


class RunnersGroupResult(GenericGroupResult[RunnersResult]):
    pass


class CollectorCallback(Callback):
    def __init__(self, cluster: object) -> None:
        self.cluster = cluster
        self.results: Dict[str, List[RunnersResult]] = {}
        """
        List of collected results for each host.
        If all commands are exited for the particular host, the number and order of results
        correspond to the number and order of the queued commands,
        for which the given callback was requested.
        """
        self._result: Optional[RunnersGroupResult] = None

    def accept(self, host: str, token: Token, result: RunnersResult) -> None:
        self.results.setdefault(host, []).append(result)
        self._result = None

    @property
    def result(self) -> RunnersGroupResult:
        """
        Merges the collected `results` into the single RunnersGroupResult instance.
        This can be useful to check merged output of the commands.

        Note that if the result of more than one command is merged,
        one SHOULD NOT use `RunnersResult.exited` and `RunnersResult.command`,
        as the exit code and command of few commands are undefined.

        :return: merged RunnersGroupResult instance.
        """
        if self._result is None:
            self._result = RunnersGroupResult(
                self.cluster,
                {host: RunnersResult.merge(results) for host, results in self.results.items()}
            )

        return self._result


RunResult = Union[RunnersGroupResult, Token]
GROUP_RUN_TYPE = TypeVar('GROUP_RUN_TYPE', bound=RunResult, covariant=True)
GROUP_SELF = TypeVar('GROUP_SELF', bound='AbstractGroup[Union[RunnersGroupResult, Token]]')


class AbstractGroup(Generic[GROUP_RUN_TYPE], ABC):
    def __init__(self, ips: Iterable[Union[str, GROUP_SELF]], cluster: object):
        from kubemarine.core.cluster import KubernetesCluster

        self.cluster = cast(KubernetesCluster, cluster)
        self.nodes: Set[str] = set()
        for ip in ips:
            if isinstance(ip, self.__class__):
                for host in ip.nodes:
                    self.nodes.add(host)
            elif isinstance(ip, str):
                self.nodes.add(ip)
            else:
                raise Exception('Unsupported connection object type')

    @abstractmethod
    def _make_group(self: GROUP_SELF, ips: Iterable[Union[str, GROUP_SELF]]) -> GROUP_SELF:
        pass

    def __eq__(self, other: object) -> bool:
        if self is other:
            return True

        if not isinstance(other, self.__class__):
            return False

        return self.nodes == other.nodes

    def __ne__(self, other: object) -> bool:
        return not self == other

    def run(self, command: str,
            warn: bool = False, hide: bool = True,
            env: Dict[str, str] = None, timeout: int = None,
            callback: Callback = None) -> GROUP_RUN_TYPE:
        caller: Optional[Dict[str, object]] = None
        if not hide:
            # fetching of the caller info should be at the earliest point
            caller = log.caller_info(self.cluster.log)
        return self._run("run", command, caller,
                         warn=warn, hide=hide, env=env, timeout=timeout, callback=callback)

    def sudo(self, command: str,
             warn: bool = False, hide: bool = True,
             env: Dict[str, str] = None, timeout: int = None,
             callback: Callback = None) -> GROUP_RUN_TYPE:
        caller: Optional[Dict[str, object]] = None
        if not hide:
            # fetching of the caller info should be at the earliest point
            caller = log.caller_info(self.cluster.log)
        return self._run("sudo", command, caller,
                         warn=warn, hide=hide, env=env, timeout=timeout, callback=callback)

    @abstractmethod
    def _run(self, do_type: str, command: str, caller: Optional[Dict[str, object]],
             **kwargs: Any) -> GROUP_RUN_TYPE:
        pass

    @abstractmethod
    def get(self, remote_file: str, local_file: str) -> None:
        pass

    def put(self, local_file: Union[io.StringIO, str], remote_file: str,
            backup: bool = False, sudo: bool = False,
            mkdir: bool = False, immutable: bool = False) -> None:
        if isinstance(local_file, io.StringIO):
            self.cluster.log.verbose("Text is being transferred to remote file \"%s\" on nodes %s"
                                     % (remote_file, list(self.nodes)))
            # This is a W/A to avoid https://github.com/paramiko/paramiko/issues/1133
            # if text contains non-ASCII characters.
            # Use the same encoding as paramiko uses, see paramiko/file.py/BufferedFile.write()

            local_stream: Union[io.BytesIO, str] = io.BytesIO(local_file.getvalue().encode('utf-8'))
            group_to_upload = self
        else:
            if not os.path.isfile(local_file):
                raise Exception(f"File {local_file} does not exist")

            self.cluster.log.verbose("Local file \"%s\" is being transferred to remote file \"%s\" on nodes %s"
                                     % (local_file, remote_file, list(self.nodes)))

            self.cluster.log.verbose('File size: %s' % os.path.getsize(local_file))
            eager_group = self.cluster.make_group(self.nodes)
            local_file_hash = eager_group.get_local_file_sha1(local_file)
            self.cluster.log.verbose('Local file hash: %s' % local_file_hash)
            remote_file_hashes = eager_group.get_remote_file_sha1(remote_file)
            self.cluster.log.verbose('Remote file hashes: %s' % remote_file_hashes)

            hosts_to_upload = []
            for remote_ip, remote_file_hash in remote_file_hashes.items():
                if remote_file_hash != local_file_hash:
                    self.cluster.log.verbose('Local and remote hashes does not match on node \'%s\' %s %s'
                                             % (remote_ip, local_file_hash, remote_file_hash))
                    hosts_to_upload.append(remote_ip)
            if not hosts_to_upload:
                self.cluster.log.verbose('Local and remote hashes are equal on all nodes, no transmission required')
                return

            local_stream = local_file
            group_to_upload = self._make_group(hosts_to_upload)

        group_to_upload._put_with_mv(local_stream, remote_file,
                                     backup=backup, sudo=sudo, mkdir=mkdir, immutable=immutable)

    def _put_with_mv(self, local_stream: Union[io.BytesIO, str], remote_file: str,
                     backup: bool, sudo: bool, mkdir: bool, immutable: bool) -> None:

        if sudo:
            self.cluster.log.verbose('A sudoer upload required')

        if backup:
            self.cluster.log.verbose('File \"%s\" backup required' % remote_file)

        if mkdir:
            self.cluster.log.verbose('A parent directory will be created')

        if immutable:
            self.cluster.log.verbose('File \"%s\" immutable set required' % remote_file)

        advanced_move_required = sudo or backup or immutable
        temp_filepath = remote_file

        if advanced_move_required:
            # for unknown reason fabric v2 can't put as sudo, and we should use WA via mv
            # also, if we need to backup the file first, then we also have to upload file to tmp first

            temp_filepath = "/tmp/%s" % uuid.uuid4().hex
            self.cluster.log.verbose("Uploading to temporary file '%s'..." % temp_filepath)

        self._put(local_stream, temp_filepath)

        if not advanced_move_required:
            return

        self.cluster.log.verbose("Moving temporary file '%s' to '%s'..." % (temp_filepath, remote_file))

        if sudo:
            mv_command = "sudo chown root:root %s && sudo mv -f %s %s" % (temp_filepath, temp_filepath, remote_file)
        else:
            mv_command = "mv -f %s %s" % (temp_filepath, remote_file)

        if backup:
            if sudo:
                mv_command = "sudo cp -f %s %s.bak$(sudo ls %s* | sudo wc -l); %s" \
                             % (remote_file, remote_file, remote_file, mv_command)
            else:
                mv_command = "cp -f %s %s.bak$(ls %s* | wc -l); %s" \
                             % (remote_file, remote_file, remote_file, mv_command)

        if mkdir:
            file_directory = "/".join(remote_file.split('/')[:-1])
            if sudo:
                mv_command = f"sudo mkdir -p {file_directory}; {mv_command}"
            else:
                mv_command = f"mkdir -p {file_directory}; {mv_command}"

        mv_command = "cmp --silent %s %s || (%s)" % (remote_file, temp_filepath, mv_command)

        if immutable:
            if sudo:
                mv_command = "sudo chattr -i %s; %s; sudo chattr +i %s" % (remote_file, mv_command, remote_file)
            else:
                mv_command = "chattr -i %s; %s; chattr +i %s" % (remote_file, mv_command, remote_file)

        self.sudo(mv_command)

    @abstractmethod
    def _put(self, local_stream: Union[io.BytesIO, str], remote_file: str) -> None:
        pass

    def _unsafe_make_runners_result(self, host_results: HostToResult) -> RunnersGroupResult:
        return RunnersGroupResult(self.cluster,
                                  {host: cast(RunnersResult, result) for host, result in host_results.items()})

    def call(self, action: Callable[..., _T], **kwargs: object) -> _T:
        func = cast(FunctionType, action)
        callable_path = "%s.%s" % (func.__module__, func.__name__)
        self.cluster.log.debug("Running %s: " % callable_path)
        result = action(self, **kwargs)
        if result is not None:
            self.cluster.log.debug(result)

        return result

    def call_batch(self: GROUP_SELF, actions: List[Callable[[GROUP_SELF], Any]]) -> None:
        for action in actions:
            self.call(action)

    def get_online_nodes(self: GROUP_SELF, online: bool) -> GROUP_SELF:
        online_hosts = [host for host, node_context in self.cluster.context['nodes'].items()
                        if node_context['access']['online'] == online]
        return self._make_group(online_hosts).intersection_group(self)

    def get_accessible_nodes(self: GROUP_SELF) -> GROUP_SELF:
        accessible = [host for host, node_context in self.cluster.context['nodes'].items()
                      if node_context['access']['accessible']]
        return self._make_group(accessible).intersection_group(self)

    def get_sudo_nodes(self: GROUP_SELF) -> GROUP_SELF:
        sudo = [host for host, node_context in self.cluster.context['nodes'].items()
                if node_context['access']['sudo'] != "No"]
        return self._make_group(sudo).intersection_group(self)

    def get_ordered_members_list(self: GROUP_SELF, apply_filter: GroupFilter = None) -> List[GROUP_SELF]:
        nodes = self.get_ordered_members_configs_list(apply_filter)
        return [self._make_group([node['connect_to']]) for node in nodes]

    def get_ordered_members_configs_list(self, apply_filter: GroupFilter = None) -> List[NodeConfig]:

        result = []
        # we have to iterate strictly in order which was defined by user in config-file
        for node in self.cluster.inventory['nodes']:
            # is iterable node from inventory is part of current NodeGroup?
            if node['connect_to'] in self.nodes:

                # apply filters
                suitable = True
                if apply_filter is not None:
                    if callable(apply_filter):
                        if not apply_filter(node):
                            suitable = False
                    else:
                        # here intentionally there is no way to filter by values in lists field,
                        # for this you need to use custom functions.
                        # Current solution implemented in this way because the filtering strategy is
                        # unclear - do I need to include when everything matches or is partial matching enough?
                        for key, value in apply_filter.items():
                            if node.get(key) is None:
                                suitable = False
                                break
                            if isinstance(value, list):
                                if node[key] not in value:
                                    suitable = False
                                    break
                            # elif should definitely be here, not if
                            elif node[key] != value:
                                suitable = False
                                break

                # if not filtered
                if suitable:
                    result.append(node)

        return result

    def get_first_member(self: GROUP_SELF, apply_filter: GroupFilter = None) -> GROUP_SELF:
        results = self.get_ordered_members_list(apply_filter)
        if not results:
            raise Exception("Failed to find first group member by the given criteria")
        return results[0]

    def get_any_member(self: GROUP_SELF, apply_filter: GroupFilter = None) -> GROUP_SELF:
        member: GROUP_SELF = random.choice(self.get_ordered_members_list(apply_filter))
        self.cluster.log.verbose(f'Selected node {member.get_host()}')
        return member

    def get_member_by_name(self: GROUP_SELF, name: str) -> GROUP_SELF:
        return self.get_first_member({"name": name})

    def new_group(self: GROUP_SELF, apply_filter: GroupFilter = None) -> GROUP_SELF:
        return self._make_group(self.get_ordered_members_list(apply_filter))

    def include_group(self: GROUP_SELF, group: GROUP_SELF) -> GROUP_SELF:
        ips = self.nodes.union(group.nodes)
        return self._make_group(ips)

    def exclude_group(self: GROUP_SELF, group: GROUP_SELF) -> GROUP_SELF:
        ips = self.nodes - group.nodes
        return self._make_group(ips)

    def intersection_group(self: GROUP_SELF, group: GROUP_SELF) -> GROUP_SELF:
        ips = self.nodes.intersection(group.nodes)
        return self._make_group(ips)

    def get_nodes_names(self) -> List[str]:
        result = []
        members = self.get_ordered_members_configs_list()
        for node in members:
            result.append(node['name'])
        return result

    def get_node_name(self) -> str:
        if len(self.nodes) != 1:
            raise Exception("Cannot get the only name from not a single node")

        return self.get_nodes_names()[0]

    def get_hosts(self) -> List[str]:
        members = self.get_ordered_members_list()
        return [node.get_host() for node in members]

    def get_host(self) -> str:
        if len(self.nodes) != 1:
            raise Exception("Cannot get the only host from not a single node")

        return next(iter(self.nodes))

    def get_config(self) -> NodeConfig:
        if len(self.nodes) != 1:
            raise Exception("Cannot get the only node config from not a single node")

        return self.get_ordered_members_configs_list()[0]

    def is_empty(self) -> bool:
        return not self.nodes

    def has_node(self, node_name: str) -> bool:
        return node_name in self.get_nodes_names()

    def get_new_nodes(self: GROUP_SELF) -> GROUP_SELF:
        return self.new_group(lambda node: 'add_node' in node['roles'])

    def get_new_nodes_or_self(self: GROUP_SELF) -> GROUP_SELF:
        new_nodes = self.get_new_nodes()
        if not new_nodes.is_empty():
            return new_nodes
        return self

    def get_nodes_for_removal(self: GROUP_SELF) -> GROUP_SELF:
        return self.new_group(lambda node: 'remove_node' in node['roles'])

    def get_changed_nodes(self: GROUP_SELF) -> GROUP_SELF:
        return self.get_new_nodes().include_group(self.get_nodes_for_removal())

    def get_unchanged_nodes(self: GROUP_SELF) -> GROUP_SELF:
        return self.exclude_group(self.get_changed_nodes())

    def get_final_nodes(self: GROUP_SELF) -> GROUP_SELF:
        return self.new_group(lambda node: 'remove_node' not in node['roles'])

    def get_initial_nodes(self: GROUP_SELF) -> GROUP_SELF:
        return self.new_group(lambda node: 'add_node' not in node['roles'])

    def nodes_amount(self) -> int:
        """
        Returns the number of nodes within a group
        :return: Integer
        """
        return len(self.nodes)

    def get_nodes_os(self) -> str:
        """
        Returns the detected operating system family for group.

        :return: Detected OS family, possible values: "debian", "rhel", "rhel8", "multiple", "unknown", "unsupported".
        """
        return self.cluster.get_os_family_for_nodes(self.nodes)

    def is_multi_os(self) -> bool:
        """
        Returns true if same group contains nodes with multiple OS families
        :return: Boolean
        """
        return self.get_nodes_os() == 'multiple'

    def get_subgroup_with_os(self: GROUP_SELF, os_family: str) -> GROUP_SELF:
        """
        Forms and returns a new group from the nodes of the original group that have a specific OS family
        :param os_family: The name of required OS family
        :return: NodeGroup
        """
        if os_family not in ['debian', 'rhel', 'rhel8']:
            raise Exception('Unsupported OS family provided')
        hosts = []
        for host in self.nodes:
            node_os_family = self.cluster.get_os_family_for_node(host)
            if node_os_family == os_family:
                hosts.append(host)
        return self._make_group(hosts)


class NodeGroup(AbstractGroup[RunnersGroupResult]):
    def _make_group(self: NodeGroup, ips: Iterable[Union[str, NodeGroup]]) -> NodeGroup:
        return NodeGroup(ips, self.cluster)

    def _make_defer(self, executor: RemoteExecutor) -> DeferredGroup:
        return DeferredGroup(self.nodes, self.cluster, executor)

    def new_defer(self, timeout: int = None) -> DeferredGroup:
        return self.new_executor(timeout).group

    def new_executor(self, timeout: int = None) -> RemoteExecutor:
        return RemoteExecutor(self, timeout=timeout)

    def get(self, remote_file: str, local_file: str) -> None:
        self._do_exec("get", remote_file, local_file)

    def _put(self, local_stream: Union[io.BytesIO, str], remote_file: str) -> None:
        self._do_exec("put", local_stream, remote_file)

    def _run(self, do_type: str, command: str, caller: Optional[Dict[str, object]],
             **kwargs: Any) -> RunnersGroupResult:
        """
        The method should be called directly from run & sudo without any extra wrappers.
        """
        do_stream = not kwargs['hide']
        if do_stream and len(self.nodes) > 1:
            raise ValueError("Streaming of output is supported only for the single node")

        if do_stream and caller is not None:
            logger = self.cluster.log
            kwargs['out_stream'] = log.LoggerWriter(logger, caller, '\t')
            kwargs['err_stream'] = log.LoggerWriter(logger, caller, '\t')

        results = self._do_exec(do_type, command, **kwargs)
        return self._unsafe_make_runners_result(results)

    def _do_exec(self, do_type: str, *args: object, **kwargs: Any) -> HostToResult:
        callback: Callback = kwargs.pop('callback', None)

        executor = RawExecutor(self.cluster, timeout=kwargs.get('timeout'))
        executor.queue(self.get_hosts(), (do_type, args, kwargs), callback=callback)
        executor.flush()

        results = {host: next(iter(results.values()))
                   for host, results in executor.get_last_results().items()}

        if any(isinstance(result, Exception) for result in results.values()):
            raise GroupResultException(NodeGroupResult(self.cluster, results))

        return results

    def wait_for_reboot(self, initial_boot_history: RunnersGroupResult, timeout: int = None) -> RunnersGroupResult:
        results = self._await_rebooted_nodes(timeout, initial_boot_history=initial_boot_history)
        if any(isinstance(result, Exception) or result == initial_boot_history.get(host)
               for host, result in results.items()):
            raise GroupResultException(NodeGroupResult(self.cluster, results))

        return self._unsafe_make_runners_result(results)

    def wait_and_get_boot_history(self, timeout: int = None) -> RunnersGroupResult:
        return self.wait_for_reboot(RunnersGroupResult(self.cluster, {}), timeout=timeout)

    def _await_rebooted_nodes(self, timeout: int = None, initial_boot_history: RunnersGroupResult = None) \
            -> HostToResult:

        executor = RawExecutor(self.cluster)
        return executor.wait_for_boot(self.get_hosts(), timeout, initial_boot_history)

    def get_local_file_sha1(self, filename: str) -> str:
        return utils.get_local_file_sha1(filename)

    def get_remote_file_sha1(self, filename: str) -> Dict[str, Optional[str]]:
        results = self.sudo("openssl sha1 %s" % filename, warn=True)
        return {host: result.stdout.split("= ")[1].strip() if result.stdout else None
                for host, result in results.items()}


class DeferredGroup(AbstractGroup[Token]):
    def __init__(self, ips: Iterable[Union[str, DeferredGroup]], cluster: object, executor: RemoteExecutor):
        super().__init__(ips, cluster)
        self._executor = executor

    @property
    def executor(self) -> RemoteExecutor:
        return self._executor

    def flush(self) -> None:
        self._executor.flush()

    def _make_group(self, ips: Iterable[Union[str, DeferredGroup]]) -> DeferredGroup:
        return DeferredGroup(ips, self.cluster, self._executor)

    def get(self, remote_file: str, local_file: str) -> None:
        self._do_queue("get", remote_file, local_file)

    def _run(self, do_type: str, command: str, caller: Optional[Dict[str, object]], **kwargs: object) -> Token:
        do_stream = not kwargs['hide']
        if do_stream:
            # To support streaming of output with use of RemoteExecutor in deferred mode, it is necessary to:
            # 1) Make sure that no two commands are executed with streaming in parallel to avoid mess in output
            # 2) Do not print output twice if error occurred.
            raise ValueError("Streaming of output is currently not supported in deferred mode")
        return self._do_queue(do_type, command, **kwargs)

    def _put(self, local_stream: Union[io.BytesIO, str], remote_file: str) -> None:
        self._do_queue("put", local_stream, remote_file)

    def _do_queue(self, do_type: str, *args: object, **kwargs: Any) -> Token:
        callback: Callback = kwargs.pop('callback', None)
        return self._executor.queue(self.get_hosts(), (do_type, args, kwargs), callback=callback)

    def include_group(self: DeferredGroup, group: DeferredGroup) -> DeferredGroup:
        self._check_same_bound_executor(group)
        return AbstractGroup.include_group(self, group)

    def exclude_group(self, group: DeferredGroup) -> DeferredGroup:
        self._check_same_bound_executor(group)
        return AbstractGroup.exclude_group(self, group)

    def intersection_group(self, group: DeferredGroup) -> DeferredGroup:
        self._check_same_bound_executor(group)
        return AbstractGroup.intersection_group(self, group)

    def _check_same_bound_executor(self, group: DeferredGroup) -> None:
        if self._executor is not group._executor:
            raise ValueError("Trying to apply set operation on deferred groups bound to different executors")


class RemoteExecutor(RawExecutor):
    def __init__(self, group: NodeGroup, connection_pool: ConnectionPool = None, timeout: int = None) -> None:
        super().__init__(group.cluster, connection_pool, timeout)
        self.group: DeferredGroup = group._make_defer(self)
        self.cluster = group.cluster

    def flush(self) -> None:
        """
        Flushes the connections' queue.
        Throws GroupException in case of any failure.

        :return: grouped tokenized results per connection.
        """
        super().flush()

        for host, results in self._last_results.items():
            if any(isinstance(result, Exception) for _, result in results.items()):
                raise RemoteGroupException(self.cluster, self._last_results)


class GroupException(Exception):
    def __init__(self, cluster: object, results: Dict[str, List[GenericResult]]):
        self.cluster = cluster
        self._results = results

    def _make_group(self, hosts: Iterable[str]) -> NodeGroup:
        return NodeGroup(hosts, self.cluster)

    def is_any_excepted(self) -> bool:
        """
        Returns true if at least one result in group is an exception

        :return: Boolean
        """
        return len(self.get_excepted_hosts_list()) > 0

    def get_excepted_hosts_list(self) -> List[str]:
        """
        Returns a list of hosts, for which the result is an exception.

        :return: List with hosts
        """
        excepted_hosts: List[str] = []
        for host, results in self._results.items():
            if any(isinstance(result, Exception) for result in results):
                excepted_hosts.append(host)
        return excepted_hosts

    def get_excepted_nodes_group(self) -> NodeGroup:
        """
        Forms and returns new NodeGroup of nodes, for which the result is an exception.

        :return: NodeGroup:
        """
        nodes_list = self.get_excepted_hosts_list()
        return self._make_group(nodes_list)

    def get_exited_hosts_list(self) -> List[str]:
        """
        Returns a list of hosts, for which the result is the completion of all commands.

        :return: List with hosts
        """
        exited_hosts: List[str] = []
        for host, results in self._results.items():
            if all(isinstance(result, RunnersResult) for result in results):
                exited_hosts.append(host)
        return exited_hosts

    def get_exited_nodes_group(self) -> NodeGroup:
        """
        Forms and returns new NodeGroup of nodes, for which the result is the completion of all commands.

        :return: NodeGroup
        """
        nodes_list = self.get_exited_hosts_list()
        return self._make_group(nodes_list)

    def __str__(self) -> str:
        # We always print output of all commands in the batch even if they are not failed,
        # for the reason that the user code might want to print output of some commands in the batch,
        # but failed to do that because of the exception.
        host_outputs = []
        for host, results in self._results.items():
            output = f"{host}:"

            # filter out transfer results and the last exception if present
            runners_results = [res for res in results if isinstance(res, RunnersResult)]
            if runners_results:
                merged_result = RunnersResult.merge(runners_results)
                output += f" code={merged_result.repr_code()}"
                repr_out = merged_result.repr_out(hide_already_printed=True)
                if repr_out:
                    output += '\n\t' + repr_out.replace('\n', '\n\t')

            # The exception may be only the last in the list. We are also sure to have at least one result per node.
            if isinstance(results[-1], Exception):
                exception = errors.wrap_kme_exception(results[-1])
                output += '\n\t' + str(exception).replace('\n', '\n\t')

            host_outputs.append(output)

        return "\n".join(host_outputs)


class GroupResultException(GroupException):
    def __init__(self, result: GenericGroupResult[GenericResult]):
        super().__init__(result.cluster, {host: [res] for host, res in result.items()})
        self.result = result


class RemoteGroupException(GroupException):
    def __init__(self, cluster: object, results: Dict[str, TokenizedResult]):
        super().__init__(cluster, {host: list(res.values()) for host, res in results.items()})
        self.results = results



# Copyright 2021-2023 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from textwrap import dedent

from kubemarine.core.action import Action
from kubemarine.core.patch import RegularPatch
from kubemarine.core.resources import DynamicResources
from kubemarine.procedures import install


class TheAction(Action):
    def __init__(self) -> None:
        super().__init__("Reconfigure Kubernetes audit policy")

    def run(self, res: DynamicResources) -> None:
        install.run_tasks(res, ['deploy.kubernetes.audit'])


class ReconfigureAuditPolicy(RegularPatch):
    def __init__(self) -> None:
        super().__init__("reconfigure_audit_policy")

    @property
    def action(self) -> Action:
        return TheAction()

    @property
    def description(self) -> str:
        return dedent(
            f"""\
            Reconfigure Kubernetes audit policy.
            To find new default target configuration, refer to kubemarine/resources/configurations/defaults.yaml
            - services.kubeadm.apiServer.extraArgs.audit-*
            - services.audit.cluster_policy
            The patch is equivalent to `kubemarine install --tasks deploy.kubernetes.audit`.
            """.rstrip()
        )



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

from kubemarine.core import log
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.plugins import manifest
from kubemarine.plugins.calico import CalicoManifestProcessor
from kubemarine.plugins.kubernetes_dashboard import get_dashboard_manifest_processor
from kubemarine.plugins.local_path_provisioner import LocalPathProvisionerManifestProcessor
from kubemarine.plugins.nginx_ingress import get_ingress_nginx_manifest_processor

MANIFEST_PROCESSOR_PROVIDERS: Dict[str, manifest.PROCESSOR_PROVIDER] = {
    "calico": CalicoManifestProcessor,
    "nginx-ingress-controller": get_ingress_nginx_manifest_processor,
    "kubernetes-dashboard": get_dashboard_manifest_processor,
    "local-path-provisioner": LocalPathProvisionerManifestProcessor,
}


def verify_inventory(inventory: dict, cluster: KubernetesCluster) -> dict:
    for plugin_name, processor_provider in MANIFEST_PROCESSOR_PROVIDERS.items():
        if not inventory["plugins"][plugin_name]["install"]:
            continue

        items = inventory['plugins'][plugin_name]['installation']['procedures']
        for i, item in enumerate(items):
            if 'python' not in item:
                continue

            config: dict = item['python']
            if config['module'] != "plugins/builtin.py" or config['method'] != "apply_yaml":
                continue

            arguments = dict(config.get('arguments', {}))
            declared_name = arguments.pop('plugin_name', None)
            if declared_name != plugin_name:
                raise Exception(f"Unexpected 'plugin_name={declared_name}' argument "
                                f"in {plugin_name!r} installation step {i}.")

            expected_args = {'plugin_name', 'original_yaml_path', 'destination_name'}
            declared_args = set(arguments.keys())
            if not declared_args.issubset(expected_args):
                raise Exception(f"Unexpected python method arguments {list(declared_args.difference(expected_args))} "
                                f"in {plugin_name!r} installation step {i}.")

            processor = processor_provider(
                cluster.log, inventory, arguments.get('original_yaml_path'), arguments.get('destination_name')
            )
            processor.validate_inventory()
            break
        else:
            cluster.log.warning(f"Invocation of plugins.builtin.apply_yaml is not found for {plugin_name!r} plugin. "
                                f"Such configuration is obsolete, and support for it may be stopped in future releases.")

    return inventory


def get_manifest_processor(logger: log.VerboseLogger, inventory: dict, plugin_name: str,
                           **arguments: str) -> manifest.Processor:
    if plugin_name not in MANIFEST_PROCESSOR_PROVIDERS:
        raise Exception(f"Manifest processor is not registered for {plugin_name!r} plugin.")

    processor_provider = MANIFEST_PROCESSOR_PROVIDERS[plugin_name]
    return processor_provider(logger, inventory, arguments.get('original_yaml_path'), arguments.get('destination_name'))


def apply_yaml(cluster: KubernetesCluster, **arguments: str) -> None:
    arguments = dict(arguments)
    plugin_name = arguments.pop('plugin_name')

    processor = get_manifest_processor(cluster.log, cluster.inventory, plugin_name, **arguments)

    manifest = processor.enrich()
    processor.apply(cluster, manifest)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module works with sysctl on remote systems.
Using this module you can generate new sysctl configs, install and apply them.
"""

import io

from kubemarine.core import utils
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup, RunnersGroupResult


def make_config(cluster: KubernetesCluster) -> str:
    """
    Converts parameters from inventory['services']['sysctl'] to a string in the format of systcl.conf.
    """
    config = ""
    if cluster.inventory['services'].get('sysctl') is not None:
        for key, value in cluster.inventory['services']['sysctl'].items():
            if isinstance(value, str):
                value = value.strip()
            if value is not None and value != '':
                value = int(value)
                if key == "kernel.pid_max":
                    required_pid_max = get_pid_max(cluster.inventory)
                    if value > 2 ** 22:
                        raise Exception(
                            "The 'kernel.pid_max' value = '%s' is greater than the maximum allowable '%s'"
                            % (value, 2 ** 22))
                    if value < required_pid_max:
                        raise Exception(
                            "The 'kernel.pid_max' value = '%s' is lower than "
                            "the minimum required for kubelet configuration = '%s'"
                            % (value, required_pid_max))
                    if value < 32768:
                        cluster.log.warning("The 'kernel.pid_max' value = '%s' is lower than "
                                            "default system value = '32768'" % value)
                config += "%s = %s\n" % (key, value)
        if not cluster.inventory['services']['sysctl'].get("kernel.pid_max"):
            pid_max = get_pid_max(cluster.inventory)
            if pid_max < 32768:
                cluster.log.warning("The 'kernel.pid_max' value = '%s' is lower than "
                                    "default system value = '32768'" % pid_max)
            if pid_max > 2**22:
                raise Exception("Calculated 'pid_max' value = '%s' is greater than the maximum allowable '%s'"
                                % (pid_max, 2**22))
            config += "%s = %s\n" % ("kernel.pid_max", pid_max)
    return config


def configure(group: NodeGroup) -> RunnersGroupResult:
    """
    Generates and uploads sysctl configuration to the group.
    The configuration will be placed in sysctl daemon directory.
    """
    config = make_config(group.cluster)
    group.sudo('rm -f /etc/sysctl.d/98-*-sysctl.conf')
    utils.dump_file(group.cluster, config, '98-kubemarine-sysctl.conf')
    group.put(io.StringIO(config), '/etc/sysctl.d/98-kubemarine-sysctl.conf', backup=True, sudo=True)
    return group.sudo('ls -la /etc/sysctl.d/98-kubemarine-sysctl.conf')


def reload(group: NodeGroup) -> RunnersGroupResult:
    """
    Reloads sysctl configuration in the specified group.
    """
    return group.sudo('sysctl -p /etc/sysctl.d/98-*-sysctl.conf')


def get_pid_max(inventory: dict) -> int:
    max_pods: int = inventory["services"]["kubeadm_kubelet"].get("maxPods", 110)
    pod_pids_limit: int = inventory["services"]["kubeadm_kubelet"].get("podPidsLimit", 4096)
    return max_pods * pod_pids_limit + 2048



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from contextlib import contextmanager
from copy import deepcopy
from typing import Optional, cast, Iterator, Type

import yaml
import ruamel.yaml

from kubemarine.core import utils, cluster as c, log, errors, static
from kubemarine.core.yaml_merger import default_merger


class DynamicResources:
    def __init__(self, context: dict, silent: bool = False) -> None:
        self.context = context
        """
        Context holding execution arguments and other auxiliary parameters which manage the execution flow.
        The context can be mutable, but only from action to action,
        and should be copied when passing to the cluster object.
        """
        self.working_context = context
        """
        Context that is directly passed to the cluster object and that holds its intermediate result.
        """

        self._silent = silent
        self._logger: Optional[log.EnhancedLogger] = None
        self._raw_inventory: Optional[dict] = None
        self._formatted_inventory: Optional[dict] = None
        self._procedure_inventory: Optional[dict] = None

        self._nodes_context: Optional[dict] = None
        """
        The nodes_context variable should hold node specific information that is not changed during Kubemarine run.
        The variable should be initialized on demand and only once.
        """

        self._cluster: Optional[c.KubernetesCluster] = None

        args: dict = context['execution_arguments']
        self.inventory_filepath: Optional[str] = args['config']
        self.procedure_inventory_filepath: Optional[str] = args.get('procedure_config')

    def logger_if_initialized(self) -> Optional[log.EnhancedLogger]:
        return self._logger

    def logger(self) -> log.EnhancedLogger:
        if self._logger is None:
            self._logger = self._create_logger()

        return self._logger

    def raw_inventory(self) -> dict:
        """Returns raw inventory, which does not preserve formatting and which should be read only."""
        if self._raw_inventory is None:
            self._load_inventory()

        return cast(dict, self._raw_inventory)

    def formatted_inventory(self) -> dict:
        """Returns raw inventory with preserved comments/quotes, to be modified and later recreated."""
        if self._formatted_inventory is None:
            self._load_inventory()

        return cast(dict, self._formatted_inventory)

    def procedure_inventory(self) -> dict:
        if self._procedure_inventory is None:
            if self.procedure_inventory_filepath:
                self._procedure_inventory = utils.load_yaml(self.procedure_inventory_filepath)
            else:
                self._procedure_inventory = {}

        return self._procedure_inventory

    def _load_inventory(self) -> None:
        if not self.inventory_filepath:
            raise Exception("Path to inventory is not defined")
        logger = self._logger
        if not self._silent:
            msg = "Loading inventory file '%s'" % self.inventory_filepath
            if logger is None:
                print(msg)
            else:
                logger.info(msg)
        try:
            data = utils.read_external(self.inventory_filepath)
            self._raw_inventory = yaml.safe_load(data)
            # load inventory as ruamel.yaml to save original structure
            self._formatted_inventory = utils.yaml_structure_preserver().load(data)
        except (yaml.YAMLError, ruamel.yaml.YAMLError) as exc:
            utils.do_fail("Failed to load inventory file", exc, logger=logger)

    def make_final_inventory(self) -> None:
        self._formatted_inventory = utils.get_final_inventory(self.cluster(), initial_inventory=self.formatted_inventory())

    def recreate_inventory(self) -> None:
        """
        Recreates initial inventory file using DynamicResources.formatted_inventory and resets all dynamic resources.

        Avoid using it directly, because cluster object and previous inventory will be lost,
        and post processing of actions may work incorrectly.
        """
        if self._formatted_inventory is None:
            return

        self._store_inventory()

        self._raw_inventory = None
        self._formatted_inventory = None
        # no need to clear _nodes_context as it should not change after cluster is reinitialized.
        # should not clear working_context as it can be inspected after execution.
        self._cluster = None

    def _store_inventory(self) -> None:
        if not self.inventory_filepath:
            raise Exception("Path to inventory is not defined")
        # replace initial inventory file with changed inventory
        with utils.open_external(self.inventory_filepath, "w+") as stream:
            utils.yaml_structure_preserver().dump(self.formatted_inventory(), stream)

    def cluster_if_initialized(self) -> Optional[c.KubernetesCluster]:
        return self._cluster

    def cluster(self) -> c.KubernetesCluster:
        """Returns already initialized cluster object or initializes new real cluster object."""
        if self._cluster is None:
            self.working_context = deepcopy(self.context)
            self._cluster = self._create_cluster(self.working_context)

        return self._cluster

    def create_deviated_cluster(self, deviated_context: dict) -> c.KubernetesCluster:
        """
        Create new cluster instance with specified deviation of context params.
        The method work should minimize work with network and avoid RW work with filesystem.
        The cluster instance should be useful to develop a patch in case enrichment procedure is changed
        and it is necessary to compare the result of old and new algorithm of enrichment.
        It should not be used in tasks.

        :param deviated_context dictionary to override context params.
        """
        sample_context = deepcopy(self.context)
        default_merger.merge(sample_context, deviated_context)
        sample_context['preserve_inventory'] = False
        args = sample_context['execution_arguments']
        args['disable_dump'] = True
        del args['ansible_inventory_location']
        return self._create_cluster(sample_context)

    @contextmanager
    def _handle_enrichment_error(self) -> Iterator[None]:
        try:
            yield
        except errors.FailException:
            raise
        except Exception as exc:
            raise errors.FailException("Failed to proceed inventory file", exc)

    def _create_cluster(self, context: dict) -> c.KubernetesCluster:
        log = self.logger()
        context['nodes'] = deepcopy(self.get_nodes_context())
        with self._handle_enrichment_error():
            cluster = self._new_cluster_instance(context)
            cluster.enrich()

        if not self._silent:
            log.debug("Inventory file loaded:")
            for role in cluster.roles:
                log.debug("  %s %i" % (role, len(cluster.ips[role])))
                for ip in cluster.ips[role]:
                    log.debug("    %s" % ip)

        args = context['execution_arguments']
        if 'ansible_inventory_location' in args:
            utils.make_ansible_inventory(args['ansible_inventory_location'], cluster)

        return cluster

    def get_nodes_context(self) -> dict:
        if self._nodes_context is None:
            with self._handle_enrichment_error():
                # temporary cluster instance to detect initial nodes context.
                light_cluster = self._new_cluster_instance(deepcopy(self.context))
                light_cluster.enrich(custom_enrichment_fns=light_cluster.get_facts_enrichment_fns())
                self._nodes_context = light_cluster.detect_nodes_context()
                light_cluster.connection_pool.close()

        return self._nodes_context

    def _new_cluster_instance(self, context: dict) -> c.KubernetesCluster:
        from kubemarine.core import flow
        cluster_class = c.KubernetesCluster
        if flow.DEFAULT_CLUSTER_OBJ is not None:
            cluster_class = flow.DEFAULT_CLUSTER_OBJ
        return cluster_class(self.raw_inventory(), context,
                             procedure_inventory=self.procedure_inventory(),
                             logger=self.logger())

    def _create_logger(self) -> log.EnhancedLogger:
        return log.init_log_from_context_args(static.GLOBALS, self.context, self.raw_inventory()).logger



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import io
import json
import uuid
import time
from typing import TypeVar, Optional

import yaml

from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup

_T = TypeVar('_T', bound='KubernetesObject')


class KubernetesObject:
    def __init__(self, cluster: KubernetesCluster, kind: str = None, name: str = None,
                 namespace: str = None, obj: dict = None) -> None:

        self._cluster = cluster
        self._reload_t: float = -1

        if not kind and not name and not namespace and not obj:
            raise RuntimeError('Not enough parameter values to construct the object')
        if obj:
            self._obj = obj
        else:
            if not kind or not name or not namespace:
                raise RuntimeError('An unsynchronized object has not enough parameters '
                                   'to be reloaded')
            self._obj = {
                "kind": kind,
                "metadata": {
                    "name": name,
                    "namespace": namespace,
                }
            }

    def __str__(self) -> str:
        return self.to_yaml()

    @property
    def uid(self) -> str:
        uid: Optional[str] = self._obj['metadata'].get('uid')
        if uid:
            return uid

        return str(uuid.uuid4())

    @property
    def kind(self) -> str:
        kind: str = self._obj['kind'].lower()
        return kind

    @property
    def namespace(self) -> str:
        namespace: str = self._obj['metadata']['namespace'].lower()
        return namespace

    @property
    def name(self) -> str:
        name: str = self._obj['metadata']['name'].lower()
        return name

    def to_json(self) -> str:
        return json.dumps(self._obj)

    def to_yaml(self) -> str:
        data: str = yaml.dump(self._obj)
        return data

    def is_reloaded(self) -> bool:
        return self._reload_t > -1

    def reload(self: _T, control_plane: NodeGroup = None, suppress_exceptions: bool = False) -> _T:
        if not control_plane:
            control_plane = self._cluster.nodes['control-plane'].get_any_member()
        cmd = f'kubectl get {self.kind} -n {self.namespace} {self.name} -o json'
        result = control_plane.sudo(cmd, warn=suppress_exceptions)
        self._cluster.log.verbose(result)
        if not result.is_any_has_code(1):
            self._obj = json.loads(result.get_simple_out())
            self._reload_t = time.time()
        return self

    def apply(self, control_plane: NodeGroup = None) -> None:
        if not control_plane:
            control_plane = self._cluster.nodes['control-plane'].get_any_member()

        json_str = self.to_json()
        obj_filename = "_".join([self.kind, self.namespace, self.name, self.uid]) + '.json'
        obj_path = f'/tmp/{obj_filename}'

        control_plane.put(io.StringIO(json_str), obj_path, sudo=True)
        control_plane.sudo(f'kubectl apply -f {obj_path} && sudo rm -f {obj_path}')
